from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union
from collections import deque


class FIFOQueue():
    def __init__(self):
        self._items = deque()


    def append(self, item):
        self._items.append(item)


    def pop(self):
        try:
            return self._items.popleft()
        except IndexError:
            pass


    def __len__(self):
        return len(self._items)


    def __contains__(self, item):
        return item in self._items


    def __iter__(self):
        yield from self._items


    def __reversed__(self):
        yield from reversed(self._items)


    def __repr__(self):
        return f"Queue({list(self._items)})"


from enum import IntEnum


class ItemConstants():
    NAME = "Name"
    NAME_LOWER = "name"
    GROUP = "Group"
    GROUPS = "Groups"
    NUM_ROWS = "NumRows"
    INTERCHANGE = "Interchange"
    INTERCHANGE_NAMES = "InterchangeNames"


class ItemTypes(IntEnum):
    NOTDEFINED = 0
    PAGE = 100
    PAGESECTION = 101
    PROPERTY = 102
    IMAGE = 103
    VIDEO = 104
    LINK = 105
    COLORSTYLE = 106
    CONTENT = 107
    CONTENT_LIST = 108
    CONTENT_PROPERTIES = 109
    IMAGE_LIST = 110
    VIDEO_LIST = 111
    EXIF = 112
    SITE = 113
    STYLE = 114
    PAGE_SECTIONS = 115
    GROUP = 116
    GROUPS = 117
    COMPONENTS = 118
    SUBJECT = 119
    OBSERVER = 120
    GRAPH = 121
    GRAPH_NODE = 122
    GRAPH_NODE_STEP = 123
    COLLECTION = 124
    COLLECTIONS = 125
    COLLECTION_SUBJECT = 126
    OBSERVER_SUBJECT = 127
    NOTIFICATION_PREFERENCE = 128 # For collection/subject
    COLLECTION_SUBJECT_SCORE = 129
    GRAPH_NODE_COLLECTION_STEP = 130
    INTERMEDIATE = 131
    NOTIFICATION_SUBJECT_PREFERENCE = 132
    NOTIFICATION_SUBJECT_PREFERENCES = 133
    ACTION = 134
    COMMAND_SETTINGS = 135
    COMMAND_DATA = 136
    COMMAND_INPUT = 137
    COMMAND_OUTPUT = 138
    COMMAND = 139
    COLLECTION_SUBJECT_SCORE_WEEK = 140
    APP_MESSAGE = 141
    SITE_MESSAGE = 142
    GOAL = 143
    ENTITY = 144
#    GRAPH_NODE_ACTION_STEP = 137

    def to_string(self, lowerTitle:bool = False):
        s:str = str(self.name)
        if (lowerTitle):
            s = s.lower().title().replace("_","")
        return s


class PropertyItemType(IntEnum):
    TYPE_STRING = 0
    TYPE_BOOLEAN = 1
    TYPE_INT = 2
    TYPE_UINT = 3
    TYPE_NUMBER = 4
    TYPE_COLOR = 5
    TYPE_DATETIME = 6
    TYPE_TEXTAREA = 7
    TYPE_EMPTY = 8 # Used for formatting the UI and therefore leave an empty space
    TYPE_LIST =  9 # List of string
    TYPE_DOUBLE = 10
    TYPE_GUID = 11
    TYPE_FLOAT = 12
    TYPE_EMAIL = 13
    TYPE_PHONE = 14
    TYPE_DATE = 15
    TYPE_URL = 16
    TYPE_STRING_LIST = 17
    TYPE_TIME = 18
    TYPE_PASSWORD = 19
    TYPE_PIXELS = 20
    TYPE_PERCENT = 21
    TYPE_HTML = 22
    TYPE_PATH = 23
    TYPE_BOOLEAN_CHECKBOX = 24
    TYPE_TEXTAREA_STRING = 25
    TYPE_STRING_BASE64 = 100
    TYPE_TEXTAREA_BASE64 = 101

    def to_string(self, lowerTitle:bool = False):
        s:str = str(self.name)
        if (lowerTitle):
            s = s.lower().title().replace("_","")
        return s


class LevelValidatedType(IntEnum):
    LEVEL_NONE = 0     # Content is not validated


class PropertySelectorType(IntEnum):
    STANDARD = 0
    AUTOCOMPLETE = 1


from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta
from ewoxcore.cache.icache_provider import ICacheProvider


class ICacheLocalProvider(ICacheProvider):
    pass



from typing import Any, Optional, Dict, Tuple, Text, List
import uuid


class UUIDUtil:
    @staticmethod
    def is_valid(id:str) -> bool:
        try:
            uuid.UUID(id)
            return True
        except ValueError:
            return False
        except:
            return False



from typing import Any, List, Dict, Text, Optional, Tuple, Union
from ewoxcore.items.item_constants import PropertySelectorType


class PropertySelector:
    def __init__(self, name:str="", type:int=0, data_type:int=int(PropertySelectorType.STANDARD)):
        self.name:str = name
        self.dataSource:str = ""
#        self.buttonItem:Item
        self.type:int = type # Used as client type
        self.dataType:int = data_type
        self.keys:List[str] = []



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from lagom import Container, Singleton, ExplicitContainer
from lagom.exceptions import InvalidDependencyDefinition, DependencyNotDefined
from ewoxcore.service.interfaces.iservice_collection import IServiceCollection
from ewoxcore.service.interfaces.iservice_provider import IServiceProvider
from ewoxcore.service.service_provider import ServiceProvider

T = TypeVar('T')
C = TypeVar('C')


class ServiceCollection(IServiceCollection):
    """ Dependency Injection Container 
    https://lagom-di.readthedocs.io/en/latest/ 
    """
    def __init__(self) -> None:
        self._container =  ExplicitContainer()
        self.add_singleton(ServiceProvider, ServiceProvider())


    def add_setting(self, type:T, impl:C) -> None:
        self.add_singleton(type, impl)


    def add_singleton(self, type:T, impl:C) -> None:
        self._container[type] = Singleton(impl)


    def add_transient(self, type:T, impl:C) -> None:
        self._container[type] = impl


    def _get_service(self, type:T) -> T:
        try:
            return self._container[type]
        except Exception as error:
            print(error)
            return None


    def build_service(self) -> IServiceProvider:
        prov:ServiceProvider = self._get_service(ServiceProvider)
        prov.add_service(self._get_service)

        return prov



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from collections.abc import Mapping
import types
from ewoxcore.service.class_registry import ClassRegistry

# def Serializable(cls):
#     class_name = cls.__name__
#     orig_init = cls.__init__

#     def __init__(self, *args, **kwargs):
#         orig_init(self, *args, **kwargs)
#         self.__type = class_name

#     def update(self, data: Union[Mapping[str, Any], Any], *,
#                 allow_new: bool = True,
#                 ignore_none: bool = False) -> Any:
#             # Mutates self by copying attributes/keys from `data`.
#             # - data can be a dict-like or another object.
#             # - allow_new=False will only update attributes that already exist on self.
#             # - ignore_none=True will skip None values from data.

#             # Accept dict-like or object
#             if isinstance(data, Mapping):
#                 src = dict(data)
#             else:
#                 # fall back to vars()/__dict__ for objects
#                 src = vars(data) if hasattr(data, "__dict__") else {}

#             for k, v in src.items():
#                 if not allow_new and not hasattr(cls, k):
#                     continue
#                 if ignore_none and v is None:
#                     continue

#                 # if isinstance(v, Mapping):
#                 #     print(f"Mapping: {k} -> {v}")
#                 if isinstance(v, types.SimpleNamespace) or hasattr(v, "__dict__"):
#                     print(f"Object: {k} -> {v}")
                
#                 # if (isinstance(v, dict)):
#                     if (hasattr(v, "__type") == True):
#                         name:str = getattr(v, "__type", "")
#                         if (name):
#                             ctor = ClassRegistry.get(name)
#                             if (ctor):
#                                 # data_v:Dict[str, Any] = dict(vars(v))
#                                 data_v:Dict[str, Any] = self.ns_to_dict(v)
#                                 v = ctor(data_v)

#                 setattr(self, k, v)

#             return self
        

#     def ns_to_dict(self, obj: Any) -> Any:
#         if isinstance(obj, types.SimpleNamespace):
#             return {k: self.ns_to_dict(v) for k, v in vars(obj).items()}
#         if isinstance(obj, dict):
#             return {k: self.ns_to_dict(v) for k, v in obj.items()}
#         if isinstance(obj, list):
#             return [self.ns_to_dict(x) for x in obj]
#         if isinstance(obj, tuple):
#             return tuple(self.ns_to_dict(x) for x in obj)

#         return obj

#     cls.__init__ = __init__
#     return cls


def Serializable(cls):
    class_name = cls.__name__
    orig_init = getattr(cls, "__init__", lambda self, *a, **k: None)

    def __init__(self, *args, **kwargs):
        orig_init(self, *args, **kwargs)
        setattr(self, "__type", class_name)


    def _ns_to_dict(self, obj: Any) -> Any:
        if isinstance(obj, types.SimpleNamespace):
            return {k: self._ns_to_dict(v) for k, v in vars(obj).items()}

        if isinstance(obj, Mapping):
            return {k: self._ns_to_dict(v) for k, v in obj.items()}
        
        if isinstance(obj, list):
            return [self._ns_to_dict(x) for x in obj]

        if isinstance(obj, tuple):
            return tuple(self._ns_to_dict(x) for x in obj)

        if hasattr(obj, "__dict__") and not isinstance(obj, (str, bytes, int, float, bool)):
            return {k: self._ns_to_dict(v) for k, v in vars(obj).items()}

        return obj


    def _merge_classes(self, data: Union[Mapping[str, Any], Any], *,
               allow_new: bool = True,
               ignore_none: bool = False) -> Any:
        if isinstance(data, Mapping):
            src = dict(data)
        else:
            src = vars(data) if hasattr(data, "__dict__") else {}

        for k, v in src.items():
            if not allow_new and not hasattr(self, k):
                continue
            if ignore_none and v is None:
                continue

            if isinstance(v, (types.SimpleNamespace, object)) and hasattr(v, "__dict__"):
                t = getattr(v, "__type", None)
                if t:
                    ctor = ClassRegistry.get(t)
                    if ctor:
                        v = ctor(self._ns_to_dict(v))

            setattr(self, k, v)

        return self

    # Attach to the class
    cls.__init__ = __init__
    cls._ns_to_dict = _ns_to_dict
    cls._merge_classes = _merge_classes

    return cls


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod

T = TypeVar('T')


class IServiceProvider(ABC):
    @abstractmethod
    def get_service(self, type:T) -> T:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from ewoxcore.service.interfaces.iservice_provider import IServiceProvider

T = TypeVar('T')


class ServiceProvider(IServiceProvider):
    def __init__(self) -> None:
        self._func:Callable[[T], T] = None

    
    def add_service(self, func:Callable[[T], T]) -> None:
        self._func = func


    def get_service(self, type:T) -> T:
        return self._func(type)



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta
from requests.models import Response
from requests.sessions import Session
from ewoxcore.client.http_client import HTTPClient
from ewoxcore.client.status_code import StatusCode
from ewoxcore.utils.json_util import JsonUtil

T = TypeVar("T")


class GraphQLClient():
    def __init__(self, url:str, token:str, request_timeout:float = 90, auth_key:str="Authorization") -> None:
        self._url:str = url
        self._token:str = token
        self._request_timeout:float = request_timeout
        self._headers:Dict[str, str] = {auth_key : self._token}


    def post(self, class_type:T, query:str, variables:str="") -> Optional[T]:
        if (not query):
            return None

        client = HTTPClient()
        session:Session = client.session(timeout=self._request_timeout)
        response:Response = None
        if (variables):
            response = session.post(url=self._url, json={"query": query, "variables": variables}, headers=self._headers)
        else:
            response = session.post(url=self._url, json={"query": query}, headers=self._headers)
        if response.status_code != int(StatusCode.Ok):
            return None

        json_data:str = response.content.decode()
        json_obj:Dict = JsonUtil.deserialize(json_data)
        query_name:str = ""
        for key, value in json_obj["data"].items():
            query_name = key
            break

        data:str = json_obj["data"][query_name][0]["data"]
        model:T = JsonUtil.deserialize_json64(class_type, data)

        return model


if __name__ == "__main__":
    client = GraphQLClient("", "")
    client.post(Any, "dudd")
    print("")


from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable
from datetime import date, datetime, timedelta
import copy
from dateutil.parser import parse
from ewoxcore.items.item_constants import PropertyItemType
from ewoxcore.items.key_value_pair import KeyValuePair
from ewoxcore.items.item_base import ItemBase
from ewoxcore.items.property_item import PropertyItem
from ewoxcore.items.item import Item
from ewoxcore.utils.uuid_util import UUIDUtil


class ItemProvider():
    @staticmethod
    def is_root(item:Item) -> bool:
        if item.parentItemId == "":
            return True
        else:
            return False


    @staticmethod
    def change_id(item:Item, itemId:str):
        item.itemId = itemId


    @staticmethod
    def get_property_item(item:Item, name:str, search_child:bool=False) -> Optional[PropertyItem]:
        if (item is None):
            return None

        prop:PropertyItem = None
        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    prop = p
                    break

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_item(child, name, search_child)
                if (prop is not None):
                    return prop

        return prop


    @staticmethod
    def get_property_item_by_id(item:Item, id:str, search_child:bool=False) -> PropertyItem:
        if (item is None):
            return None

        prop:PropertyItem = None
        for p in item.properties:
            if (p.itemId == id):
                return p

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_item_by_id(child, id, search_child)
                if (prop is not None):
                    return prop

        return prop


    @staticmethod
    def add_property(item:Item, prop:PropertyItem) -> None:
        if (item is None):
            return

        prop.idx = len(item.properties) + 1
        item.properties.append(prop)


    @staticmethod
    def add_properties(item:Item, props:List[PropertyItem]) -> None:
        if (item is None):
            return

        for prop in props:
            prop.idx = len(item.properties) + 1
            item.properties.append(prop)


    @staticmethod
    def get_subitem(item:Item, name:str, search_child:bool=False, match_item_type:int=None) -> ItemBase:
        if (item is None):
            return None

        i:ItemBase = None
        for si in item.subItems:
            if (si.name == name):
                if (match_item_type is not None):
                    if ((si.name == name) & (si.itemType == match_item_type)):
                        return si
                else:
                    return si
            if (search_child):
                i = ItemProvider.get_subitem(si, name, search_child)
                if (i is not None):
                    return i

        return i


    @staticmethod
    def get_subitem_by_id(item:Item, itemId:str, search_child:bool=False) -> ItemBase:
        if (item is None):
            return None

        i:ItemBase = None
        for si in item.subItems:
            if (si.itemId == itemId):
                return si
            if (search_child):
                i = ItemProvider.get_subitem_by_id(si, itemId, search_child)
                if (i is not None):
                    return i

        return i


    @staticmethod
    def get_subitem_by_type(item:Item, item_type:int, search_child:bool=False) -> ItemBase:
        i:ItemBase = None
        if (item is None):
            return None

        for si in item.subItems:
            if (si.itemType == item_type):
                return si
            if (search_child):
                i = ItemProvider.get_subitem_by_type(si, item_type, search_child)
                if (i is not None):
                    return i

        return i


    @staticmethod
    def remove_subitem(item:Item, item_id:str, search_child:bool=False) -> int:
        del_index:int = -1
        for index, si in enumerate(item.subItems):
            if (si.itemId == item_id):
                del_index = index
                break

            if (search_child):
                del_index = ItemProvider.remove_subitem(si, item_id, search_child)
                if (del_index != -1):
                    del item.subItems[del_index]
                    return del_index

        if (del_index != -1):
            del item.subItems[del_index]
            return -1

        return del_index


    @staticmethod
    def add_subitem(item:Item, sub_item:ItemBase) -> None:
        if (sub_item is None):
            return
        sub_item.parentItemId = item.itemId
        item.subItems.append(sub_item)


    @staticmethod
    def set_property_value(item:Item, name:str, value:str) -> None:
        if (item is None):
            return

        for p in item.properties:
            if (p.name == name):
                p.value = str(value)


    @staticmethod
    def get_property_value_string(item:Item, name:str, default_val:str="", search_child:bool=False) -> str:
        if (item is None):
            return default_val

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    return str(p.value)

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_string(child, name, default_val, search_child)
                if (prop is not None):
                    return prop

        return default_val


    @staticmethod
    def get_property_value_uuid(item:Item, name:str, search_child:bool=False) -> str:
        if (item is None):
            return ""

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    value:str = str(p.value)
                    return value if (UUIDUtil.is_valid(value)) else ""

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_uuid(child, name, search_child)
                if (prop is not None):
                    return prop

        return ""


    @staticmethod
    def get_property_value_int(item:Item, name:str, default_val:int=0, search_child:bool=False) -> int:
        if (item is None):
            return default_val

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    if (p.value == ""):
                        return default_val
                    return int(p.value)

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_int(child, name, default_val, search_child)
                if (prop is not None):
                    return prop

        return default_val


    @staticmethod
    def get_property_value_float(item:Item, name:str, default_val:float=0, search_child:bool=False) -> float:
        if (item is None):
            return default_val

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    if (p.value == ""):
                        return default_val
                    return float(p.value)

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_float(child, name, default_val, search_child)
                if (prop is not None):
                    return prop

        return default_val


    @staticmethod
    def get_property_value_datatime(item:Item, name:str, search_child:bool=False) -> datetime:
        """ DEPRECATED. Keep it for backward compatiability. """
        if (item is None):
            return None

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    if (p.value):
                        return parse(p.value)
                    else:
                        return None

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_datatime(child, name, search_child)
                if (prop is not None):
                    return prop

        return None


    @staticmethod
    def get_property_value_datetime(item:Item, name:str, search_child:bool=False) -> datetime:
        if (item is None):
            return None

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    if (p.value):
                        return parse(p.value)
                    else:
                        return None

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_datetime(child, name, search_child)
                if (prop is not None):
                    return prop

        return None


    @staticmethod
    def get_property_value_date(item:Item, name:str, search_child:bool=False) -> date:
        if (item is None):
            return None

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    if (p.value):
                        return parse(p.value).date()
                    else:
                        return None

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_date(child, name, search_child)
                if (prop is not None):
                    return prop

        return None


    @staticmethod
    def get_property_value_bool(item:Item, name:str, default_value:bool=False, search_child:bool=False) -> bool:
        if (item is None):
            return default_value

        for p in item.properties:
            if (hasattr(p, "name")):
                if (p.name == name):
                    return True if (str(p.value).lower() == "true") else False

        if (search_child):
            for child in item.subItems:
                prop = ItemProvider.get_property_value_bool(child, name, search_child)
                if (prop is not None):
                    return prop

        return False


    @staticmethod
    def create_property(name:str, label:str="", value:str="", type:int=0, is_editable:bool=True) -> PropertyItem:
        prop = PropertyItem()
        prop.name = str(name)
        prop.label = str(label)
        prop.value = str(value)
        prop.type = int(type)
        prop.isEditable = bool(is_editable)
        return prop


    @staticmethod
    def merge(old:Item, new:Item, by_name:bool=False, exclude_locked:bool=False) -> Item:
        """ Merge the new item properties into the old item. """
        if (new is None):
            return old

        for new_prop in new.properties:
            if (new_prop.value == ""):
                continue

            old_prop:PropertyItem = ItemProvider.get_property_item(old, new_prop.name)
            if (old_prop is not None):
                if (exclude_locked == True):
                    if (old_prop.isLocked == False):
                        old_prop.value = new_prop.value
                else:
                    old_prop.value = new_prop.value
            else:
                if (new_prop.label == ""):
                    new_prop.label = new_prop.name.title()
                ItemProvider.add_property(old, new_prop)

        if (by_name):
            for subitem in new.subItems:
                old_subitem:Item = ItemProvider.get_subitem(
                    item=old, 
                    name=subitem.name, 
                    search_child=False, 
                    match_item_type=subitem.itemType)
                if (old_subitem is None):
                    old.subItems.append(subitem)
                else:
                    ItemProvider.merge(old_subitem, subitem, by_name, exclude_locked=exclude_locked)
        else:
            # This is needed for situations where we have multiple items 
            # with the same name and itemType at the same level.
            for index, subitem in enumerate(new.subItems):
                if (index >= len(old.subItems)):
                    old.subItems.append(subitem)
                else:
                    old_subitem:Item = old.subItems[index]
                    ItemProvider.merge(old_subitem, subitem, by_name, exclude_locked=exclude_locked)

        return old


    @staticmethod
    def merge_props(old:Item, new:Item, include_new:bool=False) -> Item:
        """ Merge the new item properties into the old item (only existing props in old item if include_new = False)."""
        if (new is None):
            return old

        for new_prop in new.properties:
            if (new_prop.value):
                old_prop:PropertyItem = ItemProvider.get_property_item(old, new_prop.name)
                if (old_prop):
                    old_prop.value = new_prop.value
                elif (include_new):
                    if (new_prop.label == ""):
                        new_prop.label = new_prop.name.title()
                    ItemProvider.add_property(old, new_prop)

        return old


    @staticmethod
    def add_missing_props(item:Item, template:Item) -> None:
        if (template is None):
            return None

        for new_prop in template.properties:
            prop:PropertyItem = ItemProvider.get_property_item(item, new_prop.name)
            if (prop is None):
                ItemProvider.add_property(item, new_prop)


    @staticmethod
    def remove_props(item:Item, template:Item) -> None:
        if ((item is None) | (template is None)):
            return None

        for exist_prop in item.properties:
            prop:PropertyItem = ItemProvider.get_property_item(template, exist_prop.name)
            if (prop is None):
                ItemProvider.remove_property(item, exist_prop.name)


    @staticmethod
    def remove_property(item:Item, prop_name:str) -> int:
        del_index:int = -1
        for index, pi in enumerate(item.properties):
            if (pi.name == prop_name):
                del_index = index
                break

        if (del_index != -1):
            del item.properties[del_index]
            return -1

        return del_index


    @staticmethod
    def clone(item:Item, change_id:bool=True) -> Item:
        if (item is None):
            return None

        new_item:Item = Item()
        if (change_id == False):
            new_item.change_id(item.itemId)

        new_item.parentItemId = item.parentItemId
        new_item.name = item.name
        new_item.label = item.label
        new_item.itemType = item.itemType
        new_item.appType = item.appType
        if (hasattr(item, "isHidden") == False):
            setattr(item, "isHidden", False)
        else:
            new_item.isHidden = item.isHidden

        new_item.properties = copy.deepcopy(item.properties)
        new_item.subItems = ItemProvider.clone_items(item.subItems)
        new_item.relations  = copy.deepcopy(item.relations)
        new_item.externalId = item.externalId

        return new_item


    @staticmethod
    def clone_items(items:List[Item], change_id:bool=True) -> List[Item]:
        new_items:List[Item] = []
        for item in items:
            new_item:Item = ItemProvider.clone(item, change_id)
            new_items.append(new_item)

        return new_items


    @staticmethod
    def clone_property(prop:PropertyItem) -> PropertyItem:
        new = PropertyItem()
        new.name = prop.name
        new.label = prop.label
        new.itemType = prop.itemType
        new.appType = prop.appType
        new.idx = prop.idx
        new.type = prop.type
        new.value = prop.value
        new.levelValidated = prop.levelValidated
        new.isRequired = prop.isRequired
        new.labelAfter = prop.labelAfter
        new.tooltip = prop.tooltip
        new.help = prop.help
        new.isEditable = prop.isEditable
        new.isEditableDisplayable = prop.isEditableDisplayable
        new.fillRow = prop.fillRow
        new.placeholderName = prop.placeholderName
        new.isLocked = prop.isLocked
        new.selector = prop.selector

        return new


    @staticmethod
    def is_valid(item:Item) -> bool:
        if (item is None):
            return False

        for prop in item.properties:
            if ((prop.isRequired) & (prop.value == "")):
                return False
        return True


    @staticmethod
    def add_relation(item:Item, kv:KeyValuePair) -> None:
        item.relations.append(kv)


    @staticmethod
    def add_relation(item:Item, key:str, value:str) -> None:
        item.relations.append(KeyValuePair(key, value))


    @staticmethod
    def has_relation(item:Item, key:str) -> bool:
        for kv in item.relations:
            if (kv.key == key):
                return True
        return False


    @staticmethod
    def get_relation(item:Item, key:str) -> Optional[KeyValuePair]:
        for kv in item.relations:
            if (kv.key == key):
                return kv
        return None


    @staticmethod
    def get_item_by_relation(item:Item, key:str, value:str) -> Optional[Item]:
        for subItem in item.subItems:
            kv:KeyValuePair = ItemProvider.get_relation(subItem, key, value)
            if (kv):
                return subItem
        return None


    @staticmethod
    def get_item_by_relation(item:Item, key_value:KeyValuePair) -> Optional[Item]:
        return ItemProvider.get_item_by_relation(item, key_value.key, key_value.value)


    @staticmethod
    def get_subitems(item:Item, name:str, search_child:bool=False, match_item_type:int=None) -> List[ItemBase]:
        items:List[ItemBase] = []
        if (item is None):
            return []

        for si in item.subItems:
            if (si.name == name):
                if (match_item_type is not None):
                    if ((si.name == name) & (si.itemType == match_item_type)):
                        items.append(si)
                else:
                    items.append(si)
            if (search_child):
                for si in item.subItems:
                    child_items:List[ItemBase] = ItemProvider.get_subitems(si, name, search_child, match_item_type)
                    items =  items + child_items

        return items


    @staticmethod
    def is_property_number(prop:PropertyItem) -> bool:
        if ((prop.type == int(PropertyItemType.TYPE_INT)) |
            (prop.type == int(PropertyItemType.TYPE_UINT)) | 
            (prop.type == int(PropertyItemType.TYPE_FLOAT)) |
            (prop.type == int(PropertyItemType.TYPE_DOUBLE)) |
            (prop.type == int(PropertyItemType.TYPE_NUMBER)) |
            (prop.type == int(PropertyItemType.TYPE_PERCENT))):
            return True
        return False


if __name__ == "__main__":
    i1 = Item()
    i1.name = "i1"

    i1_1 = Item()
    i1_1.name = "i1_1"
    i1.add_subitem(i1_1)

    i1_2 = Item()
    i1_2.name = "i1_2"
    i1.add_subitem(i1_2)

    i2_2 = Item()
    i2_2.name = "i2_2"
    i1_2.add_subitem(i2_2)

    i1_3 = Item()
    i1_3.name = "i1_3"
    i1.add_subitem(i1_3)

    ItemProvider.remove_subitem(i1, i2_2.itemId, search_child=True)



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod

T = TypeVar('T')


class IService(ABC):
    @abstractmethod
    async def on_start(self, get_service:Callable[[Type[T]], T]) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def on_stop(self) -> None:       
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def setup(self, get_service:Callable[[Type[T]], T]):
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def stop(self):
        raise NotImplementedError("Implement inhereted method")



from typing import Any, List, Dict, Optional
from datetime import date, datetime, timedelta
import requests
from requests.sessions import Session
from requests_toolbelt.utils import dump
from urllib3.util.retry import Retry
from ewoxcore.client.timeout_adapter import TimeoutAdapter


class HTTPClient():
    def __init__(self, total:int=3, backoff_factor:float=0.5, 
                status_forcelist:List[int]=[429, 500, 502, 503, 504], 
                method_whitelist:Optional[List[str]]=None, add_logging:bool=False, 
                proxies:Dict[str,str]=None) -> None:
        self._add_logging:bool = add_logging
        self._retry:Retry = None
        self._proxies:Dict[str,str] = proxies
        if (method_whitelist is not None):
            self._retry = Retry(
                    total=total,
                    backoff_factor=backoff_factor,
                    status_forcelist=status_forcelist,
                    method_whitelist=method_whitelist
            )
        else:
            self._retry = Retry(
                    total=total,
                    backoff_factor=backoff_factor,
                    status_forcelist=status_forcelist,
            )


    def session(self, timeout:float=15) -> Session:
        """ Create a new HTTP session with the specified timeout and retry settings. """
#        adapter = HTTPAdapter(max_retries=self._retry)
        adapter = TimeoutAdapter(max_retries=self._retry, timeout=timeout)
        session = requests.Session()
        if (self._proxies):
                session.proxies.update(self._proxies)

        session.mount('http://', adapter)
        session.mount('https://', adapter)
        if (self._add_logging):
                session.hooks["response"] = [self._logging_hook]
        
        return session


    def _logging_hook(self, response, *args, **kwargs) -> None:
        data = dump.dump_all(response)
        print(data.decode('utf-8'))


#    def post(self, url:str, data:_Data, headers:Optional[Any]=None, timeout_seconds:float=15) -> Response:
#        pass



if __name__ == "__main__":
    client = HTTPClient(add_logging=True)
    session = client.session()
    print(session)
    response = session.get("https://api.openaq.org/v1/cities", params={"country": "BA"})
    print(response)

    


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
import json
from ewoxcore.configuration.configuration import Configuration
from ewoxcore.configuration.interfaces.iconfiguration_builder import IConfigurationBuilder


class ConfigurationBuilder(IConfigurationBuilder):
    def __init__(self) -> None:
        self._configuration:Configuration = Configuration()


    def build(self) -> Configuration:
        return self._configuration


    def add_json_file(self, path:str) -> IConfigurationBuilder:
        try:
            with open(path) as f:
                data:Dict = json.load(f)
                self._configuration.add(data)
        except (Exception) as error:
            pass

        return self



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod
from ewoxcore.service.interfaces.iservice_collection import IServiceCollection
from ewoxcore.service.interfaces.istartup import IStartup

T = TypeVar('T')


class IStartupInternal(IStartup, ABC):
    @abstractmethod
    def configure_defaults(self, service:IServiceCollection, get_service:Callable[[Type[T]], T]) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def on_start(self) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def on_stop(self) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def dispose(self) -> None:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
import base64
import binascii
import re

class ByteUtil:

    @staticmethod
    def to_base64(data: bytes) -> str:
        """
        Convert bytes to a Base64 encoded string.
        """
        return base64.b64encode(data).decode('utf-8')
    

    @staticmethod
    def from_base64(encoded_data: str) -> bytes:
        """
        Convert a Base64 encoded string back to bytes.
        """
        return base64.b64decode(encoded_data.encode('utf-8'), validate=True)


    @staticmethod
    def is_base64(encoded_data: str) -> bool:
        """ Check if a string is a valid Base64 encoded string.
        This method handles both standard and URL-safe Base64 formats.
        It also accepts data URIs (e.g., "data:...;base64,<data>"). """
        if not isinstance(encoded_data, str) or not encoded_data.strip():
            return False

        # Strip data URI prefix, whitespace, newlines
        enc_cleaned = encoded_data.strip().split(",")[-1].replace("\n", "").replace("\r", "")
        # Fast character filter (std + url-safe + '=')
        if not re.fullmatch(r"[A-Za-z0-9+/=_-]+", enc_cleaned):
            return False

        # Pad to multiple of 4
        enc_cleaned += "=" * (-len(enc_cleaned) % 4)

        try:
            # URL-safe decode accepts both alphabets when altchars are given
            base64.b64decode(enc_cleaned, altchars=b"-_", validate=True)
            return True
        except binascii.Error:
            return False


    @staticmethod
    def decode_base64(encoded_data: str) -> bytes:
        """ Decode a Base64 encoded string, handling both standard and URL-safe formats. """
        # keep only the data (handles "data:...;base64,<data>" too)
        encoded_data = encoded_data.strip().replace("\n", "").replace("\r", "").split(",")[-1]
        # pad to a multiple of 4 (Base64 requirement)
        encoded_data += "=" * (-len(encoded_data) % 4)
        # decode using URL-safe alphabet
        return base64.urlsafe_b64decode(encoded_data)


from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, time, datetime, timedelta, timezone
from dateutil import parser
from ewoxcore.security.models.token import Token
from ewoxcore.utils.json_util import JsonUtil
from ewoxcore.utils.dictionary_util import DictionaryUtil

T = TypeVar("T")
Y = TypeVar("Y")


class AuthProvider():
    def __init__(self, ttl_seconds:int):
        self._ttl:int = ttl_seconds


    def get_token(self, user_id:str, data:T, encrypt:Callable[[str], str]) -> str:
        now_utc:datetime = datetime.now(tz=timezone.utc)
        data_enc:str = JsonUtil.serializeJson64(data)
        token = Token(user_id, now_utc, data_enc)
        json_enc:str = JsonUtil.serialize(token)
        token_enc:str = encrypt(json_enc)

        return token_enc


    def validate(self, class_type:Y, user_id:str, token:str, decrypt:Callable[[str], str]) -> Y:
        """ Returns none if token is not valid, otherwise a model based on class type."""
        now_utc:datetime = datetime.now(tz=timezone.utc)
        json_dec:str = decrypt(token)
        token_dec:Token = JsonUtil.deserialize_gen_object(Token, json_dec)
        token_exp:datetime = parser.parse(token_dec.expireAt)
        if ((now_utc < token_exp) | 
            (user_id != token_dec.userId)):
            return None

        model:Y = JsonUtil.deserialize_json64(class_type, token_dec.data)
        return model

    """
    def convert_dict_to_string(dict_data):
        return json.dumps(dict_data)


    def convert_string_to_dict(string_data):
        return json.loads(string_data)
    
    def _add_expiry_time(self, data):
        datetime.now(tz=timezone.utc)
        data_dict = {"data": data, "expiry": str(datetime.datetime.now())}
        string_data_dict = convert_dict_to_string(data_dict)
        return string_data_dict

    def get_token(self, data:Optional[Any]=None):
        expiry_added_data = self._add_expiry_time(data)
        encrypted_data = self.fernet.encrypt(expiry_added_data)
        return encrypted_data

    def _is_token_valid(self, data_dict):
        # check if data dicts contains valid keys
        data = data_dict.get("data", None)
        expiry = data_dict.get("expiry", None)
        is_token_valid = data != None or expiry != None

        # Check if token is expired
        time_difference = get_datetime_difference(
            parser.parse(expiry), datetime.datetime.now()
        )

        is_not_expired = time_difference < float(self.ttl_in_second)

        if is_token_valid and is_not_expired:
            return True
        return False

    def get_data(self, cipher_text):
        decrypted_token = self.fernet.decrypt(cipher_text)

        if not decrypted_token:
            return "Invalid token"

        data_dict = convert_string_to_dict(decrypted_token)

        if self._is_token_valid(data_dict):
            return data_dict.get("data")
        else:
            return "Invalid token"
    """

"""
# Initialize JWT with time to live in second.
jwt = JWT(ttl_in_second=2)

data = "my awesome data"
token = jwt.get_token(data)
print(token)

print("\n")
text = jwt.get_data(token)
print(text)
"""


"""
from jose import jwt
token = jwt.encode({'key': 'value'}, 'secret', algorithm='HS256')
u'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJrZXkiOiJ2YWx1ZSJ9.FG-8UppwHaFp1LgRYQQeS6EDQF7_6-bMFegNucHjmWg'

jwt.decode(token, 'secret', algorithms=['HS256'])
{u'key': u'value'}
"""


"""
class TestModel():
    def __init__(self):
        self.name:str = "Dude"


if __name__ == "__main__":
    from ewoxcore.security.fernet_provider import FernetProvider
    key:str = FernetProvider.generate_key()
    enc_prov = FernetProvider(key)

    user_id:str = "111"
    model = TestModel()
    prov = AuthProvider(1)
    token_data:str = prov.get_token(user_id, model, enc_prov.encrypt)
    model_dec:TestModel = prov.validate(TestModel, user_id, token_data, enc_prov.decrypt)
    print("")
"""


from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from ewoxcore.decorators.serializable import Serializable

@Serializable
class PagingArgs():
    def __init__(self, data:Optional[dict[str, any]]=None) -> None:
        if data is None:
            self.search: str = ""
            self.skip: int = 0
            self.num: int = 10
            self.sortField: str = ""
            self.sortDir: str = "asc"
        else:
            self.search: str = data.get("search", "")
            self.skip: int = data.get("skip", 0)
            self.num: int = data.get("num", 10)
            self.sortField: str = data.get("sortField", "")
            self.sortDir: str = data.get("sortDir", "asc")


    def to_dict(self) -> Dict[str, Any]:
        return {
            "search": self.search,
            "skip": self.skip,
            "num": self.num,
            "sortField": self.sortField,
            "sortDir": self.sortDir,
        }



from typing import Any, Optional, Dict, Tuple, Text, List


class BooleanUtil:
    @staticmethod
    def is_bool(val:str) -> bool:
        try:
            if ((val.lower() == "true") | (val.lower() == "false")):
                return True
        except:
            return False


    @staticmethod
    def get_safe_bool(val:str) -> bool:
        if (val is None):
            return False
        return True if (str(val).lower() == "true") else False



from typing import Any, Optional, Dict, Tuple, Text, List
import re


class LanguageUtil:
    @staticmethod
    def is_valid(language_code:str, fast_check:bool=False) -> bool:
        try:
            if (fast_check):
                if (len(language_code) != 5):
                    return False
                return True
            else:
                pattern:str = "^[a-z]{2}-[A-Z]{2}$"
                res:re.Match = re.match(pattern, language_code)
                return True if (res) else False
        except ValueError:
            return False
        except:
            return False


if __name__ == "__main__":
    res:bool = LanguageUtil.is_valid("en-GB")
    print(res)
    res:bool = LanguageUtil.is_valid("e1-GB")
    print(res)
    res:bool = LanguageUtil.is_valid("en-gb")
    print(res)



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta


class ClassRegistry():
    _class_types:Dict[str, Any] = dict()


    @staticmethod
    def register(name:str, cls:Type[Any]) -> None:
        """ Register a class with a name. """
        if (name in ClassRegistry._class_types):
            return

        ClassRegistry._class_types[name] = cls


    @staticmethod
    def add(cls:Type[Any]) -> None:
        """ Add a class by its name"""
        name:str = cls.__name__
        if (name in ClassRegistry._class_types):
            return

        ClassRegistry._class_types[name] = cls


    @staticmethod
    def get(name:str) -> Type[Any]:
        """ Get a class by name. """
        if (name not in ClassRegistry._class_types):
            raise ValueError(f"Class with name '{name}' is not registered.")

        cls = ClassRegistry._class_types[name]
        if (not isinstance(cls, type)):
            raise TypeError(f"Registered handler for '{name}' is not a class.") 

        return cls


    



import pandas as pd
import numpy as np
from pandas.io.json import json_normalize


class DataFrameUtil:
    @staticmethod
    def convertToDF(dfJSON) -> pd.DataFrame:
        """ Converts the input JSON to a DataFrame. """
        return(json_normalize(dfJSON))


    @staticmethod
    def convertToJSON(df:pd.DataFrame) -> str:
        """ Converts the input DataFrame to JSON. """
        resultJSON = df.to_json(orient='records')
        return(resultJSON)


    @staticmethod
    def csvToDF(path: str, separator:str='\t', skip_rows:int=0) -> pd.DataFrame:
        """ Converts the input CSV to DataFrame. """
        return pd.read_csv(path, skiprows=skip_rows, sep=separator, comment='#')


    @staticmethod
    def convertToDict(df:pd.DataFrame) -> dict:
        """ Converts the input DataFrame to dictionary. """
        tmp_d = dict()
        for index, row in list(df.iterrows()):
            v1 = row.values[0]
            v2 = row.values[1]
            tmp_d[v1] = v2
        return tmp_d


    @staticmethod
    def remove_row(df:pd.DataFrame, id:str, column_name:str="id") -> pd.DataFrame:
        df_list:pd.DataFrame = df.loc[df[column_name] != id]
        return df_list


    @staticmethod
    def get_list(df:pd.DataFrame, order_by:str="row_idx", skip:int=0, num:int=10) -> pd.DataFrame:
        df_list:pd.DataFrame = df.nlargest(num+skip, columns=order_by).tail(num)
        return df_list


    @staticmethod
    def add_row_number(df:pd.DataFrame, column:str="row_idx") -> None:
        df[column] = np.arange(len(df))


    @staticmethod
    def search(df:pd.DataFrame, column:str, name:str, use_case:bool=False) -> pd.DataFrame:
        return df[df[column].str.contains(name, case=use_case, na=False)]


    @staticmethod
    def get_diff(input_a:pd.DataFrame, input_b:pd.DataFrame) -> pd.DataFrame:
#        return pd.concat([input_a, input_b]).drop_duplicates(keep=False)
        return input_a.compare(input_b)



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from datetime import date, datetime, timedelta
import os
import uuid
import asyncio
from ewoxcore.monitoring.ilogger import ILogger
from ewoxcore.constants.server_env import ServerEnv
from ewoxcore.service.interfaces.iservice import IService
from ewoxcore.service.interfaces.iservice_collection import IServiceCollection
from ewoxcore.service.interfaces.iservice_provider import IServiceProvider
from ewoxcore.service.interfaces.istartup_internal import IStartupInternal
from ewoxcore.monitoring.simple_logger import SimpleLogger
from ewoxcore.monitoring.signal_handler import SignalHandler
from ewoxcore.service.interfaces.istartup import IStartup
from ewoxcore.service.service_collection import ServiceCollection
from ewoxcore.service.startupbase import StartupBase
from ewoxcore.service.service import add_service

T = TypeVar('T')


class ServiceAppBase():
    def __init__(self, service_name:str="", startup:IStartup=None, service:IService=None) -> None:
        self._startup:IStartup = startup
#        self._service:IService = service if (service) else ServiceBase()
        # self._service_id:str = str(uuid.uuid4())
        self._service_name:str = service_name
        self._startup_internal:IStartupInternal = StartupBase()
        self._service_collection:IServiceCollection = ServiceCollection()
        self.get_service:Callable[[Type[T]], T] = None

#        is_debug:bool = False if (os.getenv(ServerEnv.ENVIRONMENT) == ServerEnv.PRODUCTION) else True


    async def on_start(self) -> None:
        """ Called on start up. Inheret this method. """
        pass


    async def on_stop(self) -> None:
        """ Called on stop. Inheret this method. """
        pass


    async def _on_stop(self) -> None:
        await self._startup_internal.on_stop()
        await self.on_stop()
        await self._startup_internal.dispose()


    async def _start_async(self) -> None:
        service_provider:IServiceProvider = self._service_collection.build_service()
        self.get_service = service_provider.get_service
        add_service(self.get_service)

        self._startup_internal.configure_services(self._service_name, self._service_collection, self.get_service)

        if (self._startup):
            self._startup.configure_services(self._service_collection, self.get_service)

        self._startup_internal.configure_defaults(self._service_collection, self.get_service)
        await self._startup_internal.setup_async(self.get_service)
        if (self._startup):
#            self._startup.setup(self.get_service)
            await self._startup.setup_async(self.get_service)

        signal:SignalHandler = self.get_service(SignalHandler)
        signal.set_callback(self._on_stop)

#        logger:ILogger = self.get_service(ILogger)
 #       if (self._service):
 #           await self._service.setup(self.get_service)

        await self._startup_internal.on_start()
        await self.on_start()

#        logger.info(f"Environment: {server_env}, Port: {server_port} ")
#        await asyncio.Future()

"""
    def start(self, server_port:int = 80, debug_server_port:int=0) -> None:
        uvloop.install()
        server_env:str = os.getenv(ServerEnv.ENVIRONMENT)
        if ((server_env == ServerEnv.DEVELOPMENT) & (debug_server_port != 0)):
            server_port = debug_server_port

        asyncio.run(self._start_async(server_port))
"""



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta
from ewoxcore.cache.cache import Cache
from ewoxcore.cache.icache_provider import ICacheProvider
from ewoxcore.utils.json_util import JsonUtil

T = TypeVar("T")


class CacheProvider(ICacheProvider):
    # __instance = None

    # @staticmethod
    # def getInstance():
    #     """ Static access method. """
    #     if CacheProvider.__instance == None:
    #         CacheProvider()
    #     return CacheProvider.__instance
    # def __init__(self):
    #     """ Virtually private constructor. """
    #     if CacheProvider.__instance != None:
    #         raise Exception("This class is a singleton!")
    #     else:
    #         CacheProvider.__instance = self
    #         self.cache = Cache()
    def __init__(self):
        self.cache = Cache()


    def activate_expiration(self, timer_interval:int) -> None:
        self.cache.activate_expiration(timer_interval)


    def change_use_local(self, use_local:bool) -> None:
        """ Supported only by distributed cache provider. """
        pass


    def set_cache_size(self, size:int) -> None:
        self.cache.set_cache_size(size)


    async def get(self, cacheKey:str) -> Any:
        return self.cache.get(cacheKey)


    async def insert(self, cacheKey:str, cacheValue:Any) -> None:
        self.cache.insert(cacheKey, cacheValue)


    async def remove(self, cacheKey:str) -> None:
        self.cache.remove(cacheKey)


    async def get_all_values(self) -> List[Any]:
        return self.cache.get_all_values()


    async def add(self, key:str, value:str, absolute_expiration:timedelta) -> None:
        self.cache.add(key, value, absolute_expiration=absolute_expiration)


    async def add_expire_at(self, key:str, value:str, expire_at:datetime) -> None:
        self.cache.add_expire_at(key, value, expire_at=expire_at)


    async def clear(self) -> None:
        self.cache.clear()


    async def add_ext(self, key:str, value:T, absolute_expiration:timedelta) -> None:
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.add(key, value_internal, absolute_expiration)


    async def add_expire_at_ext(self, key:str, value:T, expire_at:datetime) -> None:
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.add_expire_at(key, value_internal, expire_at)


    async def insert_ext(self, key:str, value:T) -> None:
        value_internal:str = JsonUtil.serializeJson64(value)
        await self.insert(key, value_internal)


    async def get_ext(self, class_type:T, key:str) -> T:
        value_enc:str = await self.get(key)
        return JsonUtil.deserialize_json64(class_type, value_enc)



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable
import base64


class ItemValueSerializer:   
    @staticmethod
    def serialize_base64(value:str) -> str:
        val_serialized:str = base64.standard_b64encode(value.encode()).decode()
        return val_serialized


    @staticmethod
    def deserialize_base64(value:str) -> str:
        val_deserialized:str = base64.standard_b64decode(value).decode()
        return val_deserialized


    @staticmethod
    def serialize(value:str) -> str:
        value = value.replace("'", "(!!)")
        vals:List[str] = value.split('\\')
        if (len(vals) > 1):
            value = vals[0] + "(|)" + vals[1]

        return value


    @staticmethod
    def deserialize(value:str) -> str:
        value = value.replace("(!!)", "'")
        return value.replace("(|)", "\\")



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
import os
from ewoxcore.client.list_model import ListModel
from ewoxcore.client.paging_args import PagingArgs
from ewoxcore.client.paging_model import PagingModel
from ewoxcore.constants.server_env import ServerEnv
from ewoxcore.localize.localize_provider import LocalizeProvider
from ewoxcore.monitoring.ilogger import ILogger
from ewoxcore.cache.icache_provider import ICacheProvider
from ewoxcore.cache.icache_local_provider import ICacheLocalProvider
from ewoxcore.cache.cache_provider import CacheProvider
from ewoxcore.monitoring.monitoring_event import MonitoringEvent
from ewoxcore.monitoring.monitoring_event_name import MonitoringEventName
from ewoxcore.localize.ilocalizer import ILocalizer
from ewoxcore.service.authorizer import Authorizer
from ewoxcore.service.interfaces.iauthorizer import IAuthorizer
from ewoxcore.service.interfaces.iservice_collection import IServiceCollection
from ewoxcore.service.class_registry import ClassRegistry
from ewoxcore.configuration.configuration import Configuration
from ewoxcore.configuration.interfaces.iconfiguration_builder import IConfigurationBuilder
from ewoxcore.configuration.configuration_builder import ConfigurationBuilder
from ewoxcore.service.interfaces.istartup_internal import IStartupInternal
from ewoxcore.monitoring.simple_logger import SimpleLogger
from ewoxcore.monitoring.signal_handler import SignalHandler
from ewoxcore.service.service_identifier import ServiceIdentifier

T = TypeVar('T')


class StartupBase(IStartupInternal):
    def __init__(self) -> None:
        self._get_service:Callable[[Type[T]], T]=None
        self._configuration:Configuration = None


    def configure_services(self, service_name:str, service:IServiceCollection, get_service:Callable[[Type[T]], T]=None, use_unittest:bool=False) -> None:
        service.add_singleton(IConfigurationBuilder, lambda c: ConfigurationBuilder())
        service.add_singleton(ServiceIdentifier, lambda c: ServiceIdentifier(service_name))

        env:str = os.getenv(ServerEnv.ENVIRONMENT)
        builder:IConfigurationBuilder = get_service(IConfigurationBuilder)
        builder.add_json_file(f"data/settings/appbasesettings-{env}.json")

        ClassRegistry.add(ListModel)
        ClassRegistry.add(PagingModel)
        ClassRegistry.add(PagingArgs)


    def configure_defaults(self, service:IServiceCollection, get_service:Callable[[Type[T]], T]=None) -> None:
        self._get_service = get_service
        builder:IConfigurationBuilder = get_service(IConfigurationBuilder)
        self._configuration = builder.build()

        logger:ILogger = get_service(ILogger)
        if (logger is None):
            service.add_singleton(ILogger, lambda c: SimpleLogger())
            logger = get_service(ILogger)

        logger.setup()

        service.add_singleton(SignalHandler, lambda c: SignalHandler())

        cache_local:ICacheLocalProvider = get_service(ICacheLocalProvider)
        if (cache_local is None):
            service.add_singleton(ICacheLocalProvider, lambda c: CacheProvider())

        localizer:ILocalizer = get_service(ILocalizer)
        if (localizer is None):
            service.add_singleton(ILocalizer, lambda c: LocalizeProvider())

        authorizer:IAuthorizer = get_service(IAuthorizer)
        if (authorizer is None):
            service.add_singleton(IAuthorizer, lambda c: Authorizer())


    def setup(self, get_service:Callable[[Type[T]], T]=None) -> None:
        signal:SignalHandler = get_service(SignalHandler)
        signal.setup()

        cache_local:ICacheLocalProvider = get_service(ICacheLocalProvider)
        cache_local.activate_expiration(10)


    async def setup_async(self, get_service:Callable[[Type[T]], T]) -> None:
        self.setup(get_service)


    async def on_start(self) -> None:
        pass


    async def on_stop(self) -> None:
        pass


    async def dispose(self) -> None:
        pass


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod
from ewoxcore.configuration.configuration import Configuration

T = TypeVar('T')
C = TypeVar('C')


class IConfigurationBuilder(ABC):
    @abstractmethod
    def add_json_file(self, path:str) -> T:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def build(self) -> Configuration:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Optional, Dict, Tuple, Text, List


class switch(object):
    value = None
    def __new__(class_, value):
        class_.value = value
        return True

def case(*args):
    return any((arg == switch.value for arg in args))



from typing import Any, List, Dict, Text, Optional, Tuple, Union
from datetime import datetime, timezone
import uuid

class MessageArgs():
    def __init__(self, data:str="", command:str="", correlation_id:str=None, service_name:str="", server_name:str="", send_at:datetime=datetime.now(timezone.utc)) -> None:
        self.data:str = data # The data must be encoded using JsonUtil.serializeJson64
        self.command:str = command
        self.correlationId:str = str(uuid.uuid4()) if (correlation_id is None) else correlation_id
        self.serviceName:str = service_name
        self.serverName:str = server_name
        self.sendAt:datetime = send_at



import os
import sys
import logging
import importlib
from typing import Any, List, Dict, Text, Optional, Tuple, Union
import ewoxcore.utils.file_util as file_utils


class ModuleManager:
    def __init__(self) -> None:
        self.classes = dict()


    def load_module_test(self, moduleName:str, className:str) -> Any:
        module_dir, module_file = os.path.split("modules/coachbot/actions/action_article.py")
        module_name, module_ext = os.path.splitext(module_file)
        spec = importlib.util.spec_from_file_location(module_name,"modules/coachbot/actions/action_article.py")
        module = spec.loader.load_module()
        class_ = getattr(module, className)


    def load_module(self, moduleName:str, className:str) -> Any:
        try:
            module = importlib.import_module(moduleName)          
            class_ = getattr(module, className)
            self.classes[className] = class_
        except (ImportError, NameError, SyntaxError) as err:
            logging.error("Error loading module", exc_info=True)

        return class_


    def get_class(self, className:str) -> Any:
        return self.classes[className]


    def get_class_instance(self, className:str) -> Any:
        class_ = self.classes[className]
        return class_()



import logging
from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, overload
from datetime import date, datetime, timedelta
import random
import time
import threading
from sortedcontainers import SortedDict
from ewoxcore.cache.timer_ext import TimerExt


class Cache():
    def __init__(self, size:int = 10, use_expired_queue:bool=False, timer_interval:int=60):
        self._cache:dict = {}
        self._requested_at:dict = {}
        self._expirations:SortedDict = SortedDict()
        self._cache_size:int = size
        self._lock = threading.Lock()
        self._timer:TimerExt = None
        if (use_expired_queue):
            self._expiration_start(timer_interval)

    
    def activate_expiration(self, timer_interval:int) -> None:
        """ timer interval in seconds. """
        self._expiration_start(timer_interval)


    def _expiration_start(self, timer_interval:int) -> None:
        if (self._timer is not None):
            self._timer.stop()
            self._timer = None

        self._timer = TimerExt(self._remove_expired, timer_interval, True)
        self._timer.start()


    def set_cache_size(self, size:int) -> None:
        with self._lock:
            self._cache_size = size


    def contains(self, key:str) -> bool:
        with self._lock:
            if (key in self._cache):
                return True
            return False


    def get_allocated_size(self) -> int:
        length:int = 0
        with self._lock:
            length = len(self._cache)
        return length


    def get(self, key:str) -> Any:
        """
        Retrieve an object from our cache, keeping note of the time we last 
        retrieved it. 
        """
        with self._lock:
            if key in self._cache:
                self._requested_at[key] = time.time()
                return self._cache[key]
            else:
#                raise KeyError("Not found!")
                return None


    def insert(self, key:str, value:Any) -> None:
        """
        Insert a new object into our cache. If inserting would exceed our cache size,
        then we shall make room for it by removing the least recently used. 
        """
        with self._lock:
            if len(self._cache) == self._cache_size:
                self._remove_least_recently_used()

        self._cache[key] = value
        self._requested_at[key] = time.time()


    def _remove_least_recently_used(self):
        if self._requested_at:
            # find key with the lowest timestamp.
            leastRecentlyUsedKey = min(self._requested_at, key=lambda cacheKey: self._requested_at[cacheKey])

            self._cache.pop(leastRecentlyUsedKey, None)
            self._requested_at.pop(leastRecentlyUsedKey, None)
        else:
            # otherwise choose randomly.
            randomChoice = random.choice(self._cache.keys())
            self._cache.pop(randomChoice, None)


    def remove(self, key:str) -> None:
        with self._lock:
            self._cache.pop(key, None)


    def get_all_values(self) -> List[Any]:
        values:List[Any] = []
        with self._lock:
            for key, value in self._cache.items():
                values.append(value)

        return values


    def _remove_expired(self) -> None:
        try:
            now = datetime.now()
            now_num:int = int(now.strftime('%Y%m%d%H%M%S'))
            history:datetime = datetime.now() - timedelta(days=1)
            history_num:int = int(history.strftime('%Y%m%d%H%M%S'))
            keys_expired = self._expirations.irange(history_num, now_num)
            for key in keys_expired:
                with self._lock:
                    cache_key:str = self._expirations[key]
                    self._expirations.pop(key, None)
                self.remove(cache_key)
#            self._expiration_start()
        except Exception as error:
            logging.error(f"Cache remove expired. Error: {error}")


    def add(self, key:str, value:str, absolute_expiration:timedelta) -> None:
        now_time:datetime = datetime.now()
        now = now_time + absolute_expiration
        now_num:int = int(now.strftime('%Y%m%d%H%M%S'))
        self.insert(key, value)
        with self._lock:
            self._expirations[now_num] = key


    def add_expire_at(self, key:str, value:str, expire_at:datetime) -> None:
        at_num:int = int(expire_at.strftime('%Y%m%d%H%M%S'))
        self.insert(key, value)
        with self._lock:
            self._expirations[at_num] = key


    def clear(self) -> None:
        keys:List[str] = []
        for key in self._cache.keys():
            keys.append(key)

        for key in keys:
            self.remove(key)


if __name__ == "__main__":
    cache:Cache = Cache(use_expired_queue=True, timer_interval=2)
    
    cache.add("k1", "Dude", timedelta(seconds=1))
    while True:
        print("")
    cache._remove_expired()
    time.sleep(2)
    cache._remove_expired()


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from ewoxcore.constants.global_key import GlobalKey

T = TypeVar('T')


class GlobalProvider():
    @classmethod
    def get_service(cls, type:T) -> T:
        get_service:Callable[[Type[T]], T] = globals()[GlobalKey.GetService]
        return get_service(type)


    @classmethod
    def add_service(cls, get_service:Callable[[Type[T]], T]) -> None:
        globals()[GlobalKey.GetService] = get_service



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod
from ewoxcore.service.interfaces.iservice_provider import IServiceProvider

T = TypeVar('T')
C = TypeVar('C')


class IServiceCollection(ABC):
    @abstractmethod
    def add_setting(self, type:T, impl:C) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def add_singleton(self, type:T, impl:C) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def add_transient(self, type:T, impl:C) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def build_service(self) -> IServiceProvider:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, List, Dict, Text, Optional, Tuple, Union
from datetime import datetime, timezone
from ewoxcore.message.message_args import MessageArgs

class MessageAdapter():

    @staticmethod
    def parse(data:Dict[str, Any]) -> MessageArgs:

        args:MessageArgs = MessageArgs(
            data=data.get("data", ""),
            command=data.get("command", ""),
            correlation_id=data.get("correlationId", ""),
            service_name=data.get("serviceName", ""),
            server_name=data.get("serverName", ""),
            send_at=data.get("sendAt", None),
        )

        return args



from typing import Any, Optional, Dict, Tuple, Text, List
from ewoxcore.collections.stack import Stack


class StackUtil:
    @staticmethod
    def push_range_preserve_order(source:Stack, collection:List[Any]) -> None:
        """ Push list of T onto the stack and preserver the order of the list (first one in the list is last in the stack). """
        for i in reversed(range(len(collection),-1,-1)):
            if (i < len(collection)):
                source.push(collection[i])



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from signal import SIGINT, SIGTERM, SIGKILL, signal
import asyncio

T = TypeVar('T')


class SignalHandler:
    __on_stop_callback:Callable[[None], None] = None

    def __init__(self) -> None:
        self.received_signal = False


    def setup(self) -> None:
        signal(SIGINT, self._on_handle)
        signal(SIGTERM, self._on_handle)
#        signal(SIGKILL, self._on_handle)


    def set_callback(self, on_stop:Callable[[None], None]) -> None:
        SignalHandler.__on_stop_callback = on_stop


    def _on_handle(self, signal:int, frame:Any) -> None:
        loop = asyncio.get_event_loop()
        loop.create_task(self._dispose())


    async def _dispose(self) -> None:
        if (self.received_signal):
            return

        self.received_signal = True
        if (SignalHandler.__on_stop_callback is not None):
            await SignalHandler.__on_stop_callback()



from typing import Any, List, Dict, Text, Optional, Tuple, Union
from ewoxcore.items.property_item import PropertyItem
from ewoxcore.items.item_constants import PropertyItemType
from ewoxcore.utils.url_util import UrlUtil
from ewoxcore.utils.uuid_util import UUIDUtil
from ewoxcore.utils.date_time_util import DatetimeUtil
from ewoxcore.utils.string_util import StringUtil
from ewoxcore.utils.number_util import NumberUtil
from ewoxcore.utils.dictionary_util import DictionaryUtil
from ewoxcore.utils.boolean_util import BooleanUtil


class PropertyItemAdapter():
    @staticmethod
    def convert(model:Any) -> List[PropertyItem]:
        model_dict:Dict = DictionaryUtil.convert(model)

        props:List[PropertyItem] = []
        for key, value in model_dict.items():
            if ((isinstance(value, list)) | (isinstance(value, dict))):
                continue

            value_str:str = str(value) if (value is not None) else ""

            prop:PropertyItem = PropertyItem()
            prop.name = key
            prop.value = value_str if (value is not None) else ""
            prop.type = PropertyItemAdapter.get_property_type(value_str)
            props.append(prop)

        return props


    @staticmethod
    def get_property_type(val:str) -> int:
        if (val is None):
            return int(PropertyItemType.TYPE_STRING)
        if (NumberUtil.is_integer(val)):
            return int(PropertyItemType.TYPE_INT)
        elif (NumberUtil.is_float(val)):
            return int(PropertyItemType.TYPE_FLOAT)
        elif (BooleanUtil.is_bool(val)):
            return int(PropertyItemType.TYPE_BOOLEAN)
        elif (UrlUtil.is_url(val)):
            return int(PropertyItemType.TYPE_URL)
        elif (UUIDUtil.is_valid(val)):
            return int(PropertyItemType.TYPE_GUID)
        elif (DatetimeUtil.is_time(val)):
            return int(PropertyItemType.TYPE_TIME)
        elif (DatetimeUtil.is_date(val)):
            return int(PropertyItemType.TYPE_DATETIME)
        elif (StringUtil.is_email(val)):
            return int(PropertyItemType.TYPE_EMAIL)

        return int(PropertyItemType.TYPE_STRING)



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
import pandas as pd
import os
import csv
import math
from ewoxcore.utils.dictionary_util import DictionaryUtil

T = TypeVar("T")


class CsvFileUtil:
    @staticmethod
    def split(filename:str, delimiter:str=",", row_limit:int=1000,
              output_filename="output_%s.csv", output_path="", 
              keep_headers=True) -> List[str]:
        files:List[str] = []
        with open(filename, newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=delimiter)
            current_piece:int = 1
            current_out_path:str = os.path.join(
                output_path,
                output_filename % current_piece
            )
            files.append(current_out_path)

            current_out_writer = csv.writer(open(current_out_path, "w"), delimiter=delimiter)
            current_limit:int = row_limit
            if keep_headers:
                headers = reader.next()
                current_out_writer.writerow(headers)
   
            for i, row in enumerate(reader):
                if i + 1 > current_limit:
                    current_piece += 1
                    current_limit = row_limit * current_piece
                    current_out_path:str = os.path.join(
                        output_path,
                        output_filename % current_piece
                    )
                    files.append(current_out_path)

                    current_out_writer = csv.writer(open(current_out_path, "w"), delimiter=delimiter)
                    if keep_headers:
                        current_out_writer.writerow(headers)

                current_out_writer.writerow(row)

            return files


    @staticmethod
    def split_df(df:pd.DataFrame, delimiter:str=",", row_limit:int=1000,
                output_filename:str="output_%s.csv", output_path:str="", headers:List[str]=[], add_num_zeros:int=0) -> None:
        files:List[str] = []
        low:int = 0
        high:int = row_limit
        for i in range(math.ceil(len(df) / row_limit)):
            filename:str = os.path.join(output_path, output_filename % str(i+1).zfill(add_num_zeros))
            files.append(filename)
            df[low:high].to_csv(filename, sep=delimiter, index=False)
            if (len(headers) > 0):
                df.columns = headers

            low = high
            if (high + row_limit < len(df)):
                high += row_limit
            else:
                high = len(df)

        return files


    @staticmethod
    def read_file(filename:str, callback:Callable[[Any], None], chunk_size:int=100000):
        """ Read large data file or improve read performance based on setting chunk size. """
        if (callback is None):
            return

        df_chunks = pd.read_csv(filename, chunksize=chunk_size, low_memory=False)
        for df in df_chunks:
            callback(df)
        

    @staticmethod
    def write_df(df:pd.DataFrame, delimiter:str=",", filename="output.csv") -> None:
        df.to_csv(filename, sep=delimiter, index=False)


    @staticmethod
    def write(items:List[T], filename:str, delimiter:str=",", header:List[str]=[]) -> None:
        with open(filename, "w", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=header, delimiter=delimiter)
            writer.writeheader()
            for item in items:
                item_dict = DictionaryUtil.convert(item)
                writer.writerow(item_dict)



#from items.property_selector import PropertySelector
#from items.item_base import ItemBase
from .item_base import ItemBase
from .property_selector import PropertySelector
from .item_constants import LevelValidatedType
from .item_constants import PropertyItemType
from .item_constants import ItemTypes

class PropertyItem(ItemBase):
    def __init__(self, name:str="", type:int=int(PropertyItemType.TYPE_STRING), value:str="", 
                 label:str="", is_required:bool=False, is_editable:bool=True, is_editable_displayable:bool=False,
                 selector:PropertySelector=None, tooltip:str="", parentItemId:str=""):
        super().__init__(name, label, parentItemId)
        self.idx:int = 0
        self.type:int = type
        self.value:str = value
        self.levelValidated:int = int(LevelValidatedType.LEVEL_NONE)
        self.isRequired:bool = is_required
        self.labelAfter:str = ""
        self.tooltip:str = tooltip
        self.help:str = ""
        self.isEditable:bool = is_editable
        self.isEditableDisplayable:bool = is_editable_displayable
        self.fillRow:bool = False
        self.placeholderName:str = ""
        self.isLocked:bool = False
        self.selector:PropertySelector = selector
        self.hasChanged:bool = False
        self.itemType = int(ItemTypes.PROPERTY)


    def clone(self):
        item = PropertyItem()
        item.parentItemId = self.parentItemId
        item.name = self.name
        item.itemType = self.itemType
        item.appType = self.appType
        item.idx = self.idx
        item.type = self.type
        item.value = self.value
        item.levelValidated = self.levelValidated
        item.isRequired = self.isRequired
        item.label = self.label
        item.labelAfter = self.labelAfter
        item.tooltip = self.tooltip
        item.help = self.help
        item.isEditable = self.isEditable
        item.fillRow = self.fillRow
        item.isEditableDisplayable = self.isEditableDisplayable
        item.placeholderName = self.placeholderName
        item.isLocked = self.isLocked

        return item



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod

class IMessageConsumer(ABC):
    @abstractmethod
    async def on_consume(self, command:str, correlation_id:str, json64:str, service_name:str="") -> None:
        """ Handle the consumed message. """
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Optional, Dict, Tuple, Text, List
from random import choice
from string import ascii_uppercase


class CodeUtil:
    @staticmethod
    def generate(len:int=6, to_lower:bool=True) -> str:
        val:str = ''.join(choice(ascii_uppercase) for i in range(len))
        return val.lower() if (to_lower) else val


if __name__ == "__main__":
    val:str = CodeUtil.generate()
    print(val)



from typing import Text, Any, Dict, Union, List, Type, Callable
import json
import logging
import os
import uuid
import pathlib
import shutil


def delete_directory(path:str) -> None:
    try:
        shutil.rmtree(path)
    except OSError as e:
        print ("Error: %s - %s." % (e.filename, e.strerror))


def delete(filename:str) -> None:
    try:
        os.remove(filename)
    except OSError as e:
        print ("Error: %s - %s." % (e.filename, e.strerror))
    except Exception as e:
        print("Error: %s" % e)


def create_directory(path:str) -> None:
    try:
        os.makedirs(path)
    except OSError as e:
        print ("Error: %s - %s." % (e.filename, e.strerror))


def _filename_without_prefix(file: Text) -> Text:
    """Splits of a filenames prefix until after the first ``_``."""
    return "_".join(file.split("_")[1:])


def list_directory(path: Text) -> List[Text]:
    """Returns all files and folders excluding hidden files.

    If the path points to a file, returns the file. This is a recursive
    implementation returning files in any depth of the path."""

    if not isinstance(path, str):
        raise ValueError(
            "`resource_name` must be a string type. "
            "Got `{}` instead".format(type(path))
        )

    if os.path.isfile(path):
        return [path]
    elif os.path.isdir(path):
        results = []
        for base, dirs, files in os.walk(path, followlinks=True):
            # sort files for same order across runs
            files = sorted(files, key=_filename_without_prefix)
            # add not hidden files
            good_files = filter(lambda x: not x.startswith("."), files)

            results.extend(os.path.join(base, f) for f in good_files)
            # add not hidden directories
            good_directories = filter(lambda x: not x.startswith("."), dirs)

            results.extend(os.path.join(base, f) for f in good_directories)

        return results
    else:
        raise ValueError(
            "Could not locate the resource '{}'.".format(os.path.abspath(path))
        )


def list_files(path: Text) -> List[Text]:
    """Returns all files.

    If the path points to a file, returns the file. This is a recursive
    implementation returning files in any depth of the path."""

    if not isinstance(path, str):
        raise ValueError(
            "`resource_name` must be a string type. "
            "Got `{}` instead".format(type(path))
        )

    if os.path.isfile(path):
        return [path]
    elif os.path.isdir(path):
        results = []
        for base, dirs, files in os.walk(path, followlinks=True):
            # not listing files in cache
            if '__pycache__' in base:
                continue

            # sort files for same order across runs
            files = sorted(files, key=_filename_without_prefix)
            # filter out hidden files
            good_files = filter(lambda x: not x.startswith("."), files)

            # filter out files starting with '_'
            good_files = filter(lambda x: not x.startswith("_"), good_files)

            results.extend(os.path.join(base, f) for f in good_files)
        return results
    else:
        raise ValueError(
            "Could not locate the resource '{}'.".format(os.path.abspath(path))
        )


def get_filename(path:str) -> str:
    filename:str = os.path.basename(path)
    return filename


def get_filename_no_extension(path:str) -> str:
    filename:str = pathlib.Path(path).stem
    return filename


def get_file_extension(path:str) -> str:
    filename, file_extension = os.path.splitext(path)
    return file_extension


def get_uuid_filename(filename:str) -> str:
    uuid_filename:str = str(uuid.uuid4())
    ext:str = get_file_extension(filename)
    return uuid_filename + ext


def get_file_path(path:str) -> str:
    dir_path:str = os.path.dirname(path)
    return dir_path


def get_directory_name(path:str) -> str:
    dir_name = os.path.basename(os.path.dirname(path))
    return dir_name


def get_full_filename(path:str, filename:str) -> str:
    return os.path.join(path, '') + filename


def get_safe_path(path:str) -> str:
    return os.path.join(path, '')



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from ewoxcore.monitoring.models.log_event import LogEvent
from ewoxcore.monitoring.constants.log_level import LogLevel
from ewoxcore.utils.json_util import JsonUtil

T = TypeVar('T')


class LoggingFactory():
    @staticmethod
    def create(correlation_id:str, user_id:str, company_id:str="", log_level:int=int(LogLevel.Info), message:str="", event_name:str="", event_type:str="", args:T=None) -> LogEvent:
        data_class:str = args.__class__.__name__ if (args is not None) else ""
        data:str = JsonUtil.serializeJson64(args) if (args is not None) else ""
        event:LogEvent = LogEvent(correlation_id, user_id, company_id, log_level, message, event_name, event_type, data_class, data, is_encoded=True)

        return event



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
import logging
import os
from ewoxcore.constants.server_env import ServerEnv
from ewoxcore.monitoring.ilogger import ILogger
from ewoxcore.monitoring.models.log_event import LogEvent

T = TypeVar('T')


class SimpleLogger(ILogger):
    def __init__(self) -> None:
        pass


    def setup(self) -> None:
        server_env:str = os.getenv(ServerEnv.ENVIRONMENT)
        if (server_env == ServerEnv.PRODUCTION):
            logging.basicConfig(format='%(levelname)s:%(message)s')
        else:
            logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
            logging.debug("Start logging at debug level.")


    def info(self, message:str, *args, **kwargs) -> None:
        logging.info(message, *args, **kwargs)


    def warning(self, message:str, *args) -> None:
        logging.warning(message, *args)


    def error(self, message:str, *args) -> None:
        if (args):
            logging.error(message, *args, exc_info=True)
        else:
            logging.error(message, *args, exc_info=False)


    def debug(self, message:str, *args) -> None:
        logging.debug(message, *args)


    def info_ext(self, event:LogEvent) -> None:
        logging.info(event.message)


    def warning_ext(self, event:LogEvent) -> None:
        logging.warning(event.message)


    def error_ext(self, event:LogEvent) -> None:
        logging.error(event.message, exc_info=False)


    def debug_ext(self, event:LogEvent) -> None:
        logging.debug(event.message)


    def log_scoped(self, level: str, message: str, args: Optional[T] = None) -> None:
        pass



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from i18n import config
from i18n import translations
from i18n import resource_loader
from i18n.translator import t
from ewoxcore.localize.ilocalizer import ILocalizer


class LocalizeProvider(ILocalizer):
    def __init__(self) -> None:
        self._languages:dict = {}
        self._languages["en-GB"] = "en"
        self._path:str = "./data/translations"


    def register(self, language_code:str, language_two_letter:str) -> None:
        self._languages[language_code] = language_two_letter


    def _get_language(self, language_code:str) -> str:
        if language_code in self._languages:
            return self._languages[language_code]
        return "en"

    
    def setup(self, path:str="", default_language:str="en-GB") -> None:
        load_path:str = path if (path != "") else self._path
        config.set("file_format", "json")
        config.set("load_path", [load_path])
        config.set("filename_format", "{locale}.{format}")
        config.set('skip_locale_root_data', True)
        config.set("locale", self._get_language(default_language))
        resource_loader.init_json_loader()
        translations.has("foo") # to start loading files to memory.


    def get(self, key:str, language_code:str="en-GB") -> str:
        text:str = t(key, locale=self._get_language(language_code))
        return text



from typing import Any, Optional, Dict, Tuple, Text, List
import re
import urllib
import hashlib
import shortuuid
import validators


class UrlUtil:
    def __init__(self):
        pass
    

    @staticmethod
    def get_safe_url(name:str, with_short_id:bool = True) -> str:
        name = name.replace(" ","-").replace("~","-").lower()
        name = re.sub('[^A-Za-z0-9\\-]+', '', name)
#        url:str = urllib.parse.quote(name).replace("~","-")
        url:str = urllib.parse.quote(name)
        if (with_short_id):
            short_id:str = shortuuid.uuid()
            url += "-" + short_id
        return url


    @staticmethod
    def get_hash_code(url:str) -> str:
        return hashlib.md5(url.encode('utf-8')).hexdigest()


    @staticmethod
    def safe_trailing_slash(url:str) -> str:
        url_last_ch:str = url[len(url)-1:]
        if (url_last_ch != "/"):
            url += "/"
        return url


    @staticmethod
    def get_query_params(url:str) -> Dict:
        if ("?" in url):
            url = url.split("?")[1]
        params = urllib.parse.parse_qsl(url)

        params_dict:Dict = dict(params)       
        return params_dict


    @staticmethod
    def is_url(url:str) -> bool:
        try:
            res:bool = validators.url(url)
            return True if (res) else False
        except:
            return False


if __name__ == "__main__":
    url:str = "userId=234&EI=1&transport=websocket"
    params = UrlUtil.get_query_params(url)
    print(params)
    
    url:str = "http://test.com?userId=234&EI=1&transport=websocket"
    params = UrlUtil.get_query_params(url)
    print(params)



from typing import Any, List, Optional, Dict, Tuple, Text
from datetime import date, time, datetime, timedelta, timezone
from zoneinfo import ZoneInfo
from dateutil import tz, parser
from ewoxcore.constants.week_day import WeekDay


class DatetimeUtil:
    @staticmethod
    def is_time(input:str) -> bool:
        import time
        try:
            time.strptime(input, '%H:%M')
            return True
        except ValueError:
            return False
   

    @staticmethod
    def is_date(input:str, fuzzy:bool=False) -> bool:
        """
        Return whether the string can be interpreted as a date.

        :param input: str, input to check for date
        :param fuzzy: bool, ignore unknown tokens in string if True
        """
        try: 
            parser.parse(input, fuzzy=fuzzy)
            return True
        except ValueError:
            return False
        except:
            return False


    @staticmethod
    def get_safe_date_string(val:date) -> str:
        try:
            if ((val is None) | (val == "")):
                return ""
            if (str(val) == "NaT"):
                return ""
            # return str(val)
            date_val:date = DatetimeUtil.get_date(str(val))
            return date_val.isoformat()
        except Exception as error:
            return ""


    @staticmethod
    def get_date(date_s:str) -> date:
        date_str:str = ""
        try:
            date_d:datetime = parser.parse(date_s, fuzzy=False)
            date_str = date_d.strftime("%Y-%m-%d")
        except Exception as error:
            try:
                date_str = datetime.strptime(date_s, '%Y-%m-%d %H:%M:%S').strftime("%Y-%m-%d")
            except Exception as error:
                try:
                    date_str = datetime.strptime(date_s, '%Y-%m-%dT%H:%M:%S.%f000').strftime("%Y-%m-%d")
                except Exception as error:
                    date_str = datetime.strptime(date_s, '%Y-%m-%d').strftime("%Y-%m-%d")

        tmp_date_d:datetime = datetime.strptime(date_str, '%Y-%m-%d')
        date_d:date = date(tmp_date_d.year, tmp_date_d.month, tmp_date_d.day)

        return date_d


    @staticmethod
    def get_date_time(date_s:str) -> Optional[datetime]:
        date_d:datetime = None
        try:
            date_d = parser.parse(date_s, fuzzy=False)
        except Exception as error:
            return None
        return date_d


    @staticmethod
    def get_utc_isoformat() -> str:        
        utc_date:str = datetime.now(tz=timezone.utc).isoformat() + "Z"
        return utc_date


    @staticmethod
    def get_next_weekday(input:date=date.today()):
        weekday:datetime = input + timedelta(days=1)
        if weekday.isoweekday() in set((6, 7)):
            weekday += timedelta(days=-weekday.isoweekday() + 8)
        return weekday


    @staticmethod
    def get_next_weekend_day(input:date=date.today()):
        if (input.weekday() == 5):
            input += timedelta(days=1)
            return input
        weekend_date = input + timedelta((12 - input.weekday()) % 7)
        return weekend_date


    @staticmethod
    def convert_utc(input:datetime) -> datetime:
        input_utc:datetime = input.astimezone(timezone.utc)
        return input_utc


    @staticmethod
    def get_week_number(input:datetime) -> int:
        week:int = input.isocalendar().week
        return week


    @staticmethod
    def get_week_number(input:date) -> int:
        week:int = input.isocalendar().week
        return week


    @staticmethod
    def get_previous_week_number(input:datetime) -> int:
        week:int = input.isocalendar().week
        week -= 1
        if (week == 0):
            week = 52
        return week


    @staticmethod
    def get_next_day_from_days(week_day:int, days:List[int]) -> int:
        """ Expects a sorted list of days"""
        for day in days:
            if (day > week_day):
                return day

        return -1


    @staticmethod
    def get_end_of_today(delta:int=0) -> datetime:
        end_of_today:datetime = datetime.now() + timedelta(days=delta)
        end_of_today = datetime(
            year=end_of_today.year, month=end_of_today.month, day=end_of_today.day, 
            hour=23, minute=59, second=59)
        return end_of_today


    @staticmethod
    def get_next_time_slot(input:datetime, days:List[int], now:datetime, additional_delta:timedelta=timedelta(minutes=1)) -> datetime:
        if (len(days) == 0):
            return input

        week_days:List[int] = [x if (x != 0) else 8 for x in days]
        week_days.sort()

        days.sort()
        day_chosen:int = -1
        today_idx:int = -1
        now_utc:datetime = DatetimeUtil.convert_utc(now+additional_delta)
        iso_weekday:int = now_utc.isoweekday()
        input_utc:datetime = DatetimeUtil.convert_utc(input)
        for index, day in enumerate(week_days):
            if (day >= input.isoweekday()):
                if (day == 8):
                    if (iso_weekday == 0):
                        today_idx = index
                elif (day == iso_weekday):
                    today_idx = index

                # we add 10 minutes by default to ensure we get a datetime in the future for scheduling.
                if (input_utc > now_utc):
                    day_chosen = day
                    break

        if (day_chosen == -1):
            if ((today_idx > -1) & (len(week_days) > today_idx+1)):
                day_chosen = week_days[today_idx+1]
        
        week_offset:timedelta = None
        if ((day_chosen == -1) & (len(days) == 1)):
            week_offset = timedelta(days=7)
        if (day_chosen == -1):
            day_chosen = DatetimeUtil.get_next_day_from_days(iso_weekday, week_days)

        day_today:int = input.isoweekday()
        if (week_offset is not None):
            input += week_offset
        elif (day_chosen >= day_today):
            input += timedelta(day_chosen - day_today)
        else:
            diff:int = 8 - day_today
            input += timedelta(diff)

        return input


    @staticmethod
    def get_week_dates(today:datetime=datetime.today()) -> List[datetime]:
        weekday:int = today.isoweekday()
        start:datetime = today - timedelta(days=weekday)
        dates:List[datetime] = [start + timedelta(days=i) for i in range(7)]
        return dates


    @staticmethod
    def get_week_dates_as_string(today:datetime=datetime.today()) -> List[str]:
        weekday:int = today.isoweekday()
        start:datetime = today - timedelta(days=weekday)
        dates:List[str] = [str((start + timedelta(days=i)).date()) for i in range(7)]
        return dates


    @staticmethod
    def get_previous_week_range(value:date) -> Tuple[date, date]:
        # Minus 1 day as a calendar week start day is a Sunday.
        weekday:int = value.weekday()
        start_delta:timedelta = timedelta(days=weekday, weeks=1) - timedelta(days=-1)
        if (weekday == int(WeekDay.Sunday)):
            start_delta = timedelta(days=7) 
        start_date:date = value - start_delta
        end_date:date = date(year=start_date.year, month=start_date.month, day=start_date.day)
        end_date = end_date + timedelta(days=6)

        return start_date, end_date


    @staticmethod
    def get_week_range() -> Tuple[date, date]:
        start_date:date = datetime.today() - timedelta(days=datetime.today().isoweekday() % 7)
        end_date:date = start_date + timedelta(days=6)

        return start_date, end_date


    @staticmethod
    def convert_delta_to_hours(delta:timedelta) -> int:
        total_seconds:float = delta.total_seconds()
        hours = int(total_seconds / 3600)
        return hours

    
    @staticmethod
    def get_iso_string(value:str) -> str:
        dt_iso:str = DatetimeUtil.get_safe_date_string(value)
        if (len(dt_iso) < 19):
            dt_iso += "T00:00:00"
        else:
            dt_iso = dt_iso[:19]
            dt_iso += "Z" if (dt_iso[-1] != "Z") else ""

        return dt_iso


    @staticmethod
    def get_utc_string_or_none(value: Optional[str]) -> Optional[str]:
        if not value:
            return None

        try:
            dt = parser.parse(value)
            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=datetime.now().astimezone().tzinfo)

            dt_iso:str = dt.astimezone(timezone.utc).isoformat(timespec="milliseconds").replace("+00:00", "Z")

            return dt_iso
        except Exception:
            return None


if __name__ == "__main__":
    DatetimeUtil.get_week_range()
    input:date = date(year=2023, month=3, day=12)
    from_date, to_date = DatetimeUtil.get_previous_week_range(input)
    print(f"Input: {input}, Start: {from_date}, To: {to_date} ")

    input:date = date(year=2023, month=2, day=25)
    from_date, to_date = DatetimeUtil.get_previous_week_range(input)
    print(f"Input: {input}, Start: {from_date}, To: {to_date} ")

    is_time:bool = DatetimeUtil.is_time("12:23")
    week_dates = DatetimeUtil.get_week_dates_as_string()
    week_dates = DatetimeUtil.get_week_dates()
    week:int = DatetimeUtil.get_week_number(datetime.today())
    required_output:datetime = datetime(2022,4,30, 12,36,0)

    days:List[int] = [1,2,3,4,5]

    time_zone:str = "Europe/Copenhagen"
    time_of_day:int=16
    time_of_day_minute:int=53
    scheduled_at:datetime = datetime.combine(datetime.now(),
                                time(hour=time_of_day, minute=time_of_day_minute, second=0, tzinfo=ZoneInfo(time_zone)))
    next_scheduled_at:datetime = DatetimeUtil.get_next_time_slot(scheduled_at, 
                                                    days, 
                                                    now=datetime.now(), 
                                                    additional_delta=timedelta(seconds=1))

    now_utc:datetime = DatetimeUtil.convert_utc(datetime.now())
    next_scheduled_at_utc:datetime = DatetimeUtil.convert_utc(next_scheduled_at)
#    if (next_scheduled_at <= scheduled_at):
    if (next_scheduled_at_utc <= now_utc):
        print("")

    now:datetime = datetime(2023,2,22, 8,36,0)
    tzinfo = ZoneInfo("Australia/Adelaide")
    time_local = time(hour=8, minute=0, second=0, tzinfo=ZoneInfo("Australia/Adelaide"))
    scheduled_at:datetime = datetime.combine(now, time_local)
#    scheduled_at -= timedelta(days=1)
    print(str(scheduled_at))
    output:datetime = DatetimeUtil.get_next_time_slot(scheduled_at, days, now)
    print(str(output))
    if (output <= scheduled_at):
        output += timedelta(days=1)
    print(str(output))

    now:datetime = datetime(2022,4,29, 14,36,0)
    days:List[int] = [0,1,2,3,4,5,6]
    test_day:datetime = datetime(2022,4,29, 12,36,0)
    output:datetime = DatetimeUtil.get_next_time_slot(test_day, days, now)
    print(str(output))

    res:bool = required_output == output


    days:List[int] = [1, 2, 3, 4, 5] #[1, 2, 0]
    test_day:datetime = datetime(2022,4,3, 12,36,0)
    output:datetime = DatetimeUtil.get_next_time_slot(test_day, days, now)
    print(str(output))

#        scheduled_at:datetime = datetime.combine(self.datetime_provider.now(),
#                                    time(hour=time_of_day, minute=time_of_day_minute, second=0, tzinfo=ZoneInfo(time_zone)))
    time_zone:str = "Europe/Copenhagen"
    time_zone_hosting:str = "Europe/Dublin"
    time_of_day:int = 15
    time_of_day_minute:int = 38

    time_zone:str = "Europe/Kiev"
    tzinfo=ZoneInfo(time_zone_hosting)
    now_hosting:datetime = datetime.now(tzinfo)
    scheduled_at:datetime = datetime.combine(datetime.now(), #now_hosting,
                                    time(hour=time_of_day, minute=time_of_day_minute, second=0, tzinfo=ZoneInfo(time_zone)))
    print(str(now_hosting))
    print(str(scheduled_at))

    scheduled_at_utc:datetime = DatetimeUtil.convert_utc(scheduled_at)
    print(str(scheduled_at_utc))


    tzinfo=tz.gettz(time_zone_hosting)
    now_hosting:datetime = datetime.now(tzinfo)
    scheduled_at:datetime = datetime.combine(now_hosting,
                                    time(hour=time_of_day, minute=time_of_day_minute, second=0, tzinfo=tz.gettz(time_zone)))
    print(str(scheduled_at))
    days:List[int] = [0, 1, 2, 3, 4, 5, 6] #[1, 2, 0]
    output:datetime = DatetimeUtil.get_next_time_slot(scheduled_at, days, additional_delta=timedelta(seconds=1), now=now)
    print(str(output))

    scheduled_at:datetime = datetime.combine(datetime.now(),
                                    time(hour=time_of_day, minute=time_of_day_minute, second=0, tzinfo=tz.gettz(time_zone)))
    days:List[int] = [0, 1, 2, 3, 4, 5, 6] #[1, 2, 0]
    output:datetime = DatetimeUtil.get_next_time_slot(scheduled_at, days, now)
    print(str(scheduled_at))
    print(str(output))


    days:List[int] = [3, 4, 5] #[1, 2, 0]
    test_day:datetime = datetime(2022,3,24,11,15,0)
    output:datetime = DatetimeUtil.get_next_time_slot(test_day, days, now)
    print(str(output))

#    test_day:datetime = datetime(2022,2,22,11,0,0)
#    new_day:datetime = DatetimeUtil.get_next_day(test_day, days)

    next_weekend:datetime = DatetimeUtil.get_next_weekend_day(date(2022, 1, 1))
    print(next_weekend)
    next_weekend:datetime = DatetimeUtil.get_next_weekend_day(date(2022, 1, 2))
    print(next_weekend)
    
    next_weekday:datetime = DatetimeUtil.get_next_weekday(date(2022, 1, 1))
    print(next_weekday)


from typing import Any, Optional, Dict, Tuple, Text, List
from cryptography.fernet import Fernet, InvalidToken


class FernetProvider():
    def __init__(self, key:str):
        self._fernet = Fernet(key=key.encode("utf-8"))


    @staticmethod
    def generate_key() -> str:
        key:bytes = Fernet.generate_key()
        return key.decode("utf-8")


    def encrypt(self, value:str) -> str:
        enc:bytes = self._fernet.encrypt(value.encode("utf-8"))
        return enc.decode("utf-8")


    def decrypt(self, value:str) -> Optional[str]:
        dec:bytes = None
        
        try:
            dec = self._fernet.decrypt(value.encode("utf-8"))
        except InvalidToken:
            return None

        return dec.decode("utf-8")


if __name__ == "__main__":
    key:str = FernetProvider.generate_key()
    prov = FernetProvider(key)
    enc:str = prov.encrypt("Dude has a car")
    dec:str = prov.decrypt(enc)
    print(dec)




from typing import Any
from collections.abc import Mapping
from argparse import Namespace
from types import SimpleNamespace
import inspect


class DictionaryUtil:
    @staticmethod
    def get(dict:dict[str, Any], name:str, default:Any) -> Any:
        return dict[name] if (name in dict) else default


    # @staticmethod
    # def to_dict(obj) -> dict[str, Any]:
    #     d = {}
    #     for name in dir(obj):
    #         value = getattr(obj, name)
    #         if not name.startswith('__') and not inspect.ismethod(value):
    #             d[name] = value
    #     return d


    @staticmethod
    def to_dict(obj: Any) -> Any:
        # Primitives pass through
        if obj is None or isinstance(obj, (bool, int, float, str)):
            return obj

        # Containers: recurse
        if isinstance(obj, dict):
            return {k: DictionaryUtil.to_dict(v) for k, v in obj.items()}

        if isinstance(obj, (list, tuple, set)):
            t = type(obj)
            return t(DictionaryUtil.to_dict(v) for v in obj)

        # If the object exposes its own to_dict, use it
        if hasattr(obj, "to_dict") and callable(getattr(obj, "to_dict")):
            return obj.to_dict()

        # Generic Python object: walk its instance attributes
        if hasattr(obj, "__dict__"):
            return {k: DictionaryUtil.to_dict(v) for k, v in obj.__dict__.items() if not k.startswith("_")}

        return obj


    @staticmethod
    def convert(model:Any) -> dict[str, Any]:
        d = dict((name, getattr(model, name)) for name in dir(model) if not name.startswith('__') and not inspect.ismethod(name))

        return d


    @staticmethod
    def normalize(obj: Any) -> dict[str, Any]:
        if isinstance(obj, (Namespace, SimpleNamespace)):
            return {k: DictionaryUtil.normalize(v) for k, v in vars(obj).items()}

        if isinstance(obj, Mapping):
            return {k: DictionaryUtil.normalize(v) for k, v in obj.items()}

        if isinstance(obj, list):
            return [DictionaryUtil.normalize(v) for v in obj]
        if isinstance(obj, tuple):
            return tuple(DictionaryUtil.normalize(v) for v in obj)

        if isinstance(obj, set):
            return [DictionaryUtil.normalize(v) for v in obj]

        return obj



from typing import Any, List, Dict, Text, Optional, Tuple, Union
import json
from ewoxcore.items.adapters.property_item_adapter import PropertyItemAdapter
from ewoxcore.items.item import Item
from ewoxcore.items.property_item import PropertyItem
from ewoxcore.items.item_provider import ItemProvider
from ewoxcore.utils.json_util import JsonUtil
from ewoxcore.utils.string_util import StringUtil


class JsonItemAdapter():
    @staticmethod
    def convert(json_decoded:str, item:Item=None) -> Item:
        json_dict:Dict = json.loads(json_decoded)

        if (item is None):
            item = Item()

        provider = ItemProvider()
        for key, value in json_dict.items():
            if ((isinstance(value, list)) | (isinstance(value, dict))):
                continue
            if (value is None):
                continue

            prop_name:str = str(key).lower()
            prop:PropertyItem = provider.get_property_item(item, prop_name)
            if (prop is None):
                prop = PropertyItem()
                prop.name = prop_name
                prop.label = StringUtil.title_capitalized(str(key))
                prop.value = str(value)
                prop.type = PropertyItemAdapter.get_property_type(value)
                item.add_property(prop)
            elif (value != ""):
                prop.value = str(value)
                prop.type = PropertyItemAdapter.get_property_type(value)

            internal_prop_name:str = "_"+ str(key).lower()
            internal_prop:PropertyItem = provider.get_property_item(item, internal_prop_name)
            if (internal_prop is not None):
                internal_prop.value = str(value)

        return item


    @staticmethod
    def convert_to_json(item:Item) -> str:
        if (item is None):
            return "{}"

        json_dict:Dict = dict()
        for prop in item.properties:
            if (prop.name):
                json_dict[prop.name] = prop.value

        json_str:str = JsonUtil.serialize(json_dict, keep_object=False)
        return json_str


    @staticmethod
    def convert_to_json64(item:Item) -> str:
        json_dict:Dict = dict()
        for prop in item.properties:
            # We exclude internal properties starting with _
            if (prop.name.startswith("_") == False):
                json_dict[prop.name] = prop.value

        json_str:str = JsonUtil.serializeJson64(json_dict, keep_object=False)
        return json_str



import uuid

class ItemBase:
    def __init__(self, name:str="", label:str="", parentItemId:str=""):
        self.itemId:str = str(uuid.uuid4())
        self.parentItemId:str = parentItemId
        self.name:str = name
        self.label:str = label
        self.itemType:int = 0
        self.appType:int = 0
        




# TODO: Split into different classes (Like: Person, Article, WebPage etc.)
class InterchangeProperty():
    Name = "name"

    Url = "url"
    UseProxy = "useproxy"
    Html = "html"
    Text = "text"
    Title = "title"
    Subtitle = "subtitle"
    Tag = "tag"
    Author = "author"
    Credits = "credits"
    Source = "source"

    Query = "query"

    Protocol = "protocol"
    IpAddress = "ipaddress"
    Host = "host"
    Token = "token"
    Username = "username"
    Password = "password"

    Path = "path"
    Filename = "filename"

    Image = "image"
    Video = "video"

    Date = "date"
    DateTime = "datetime"
    Time = "time"

    Email = "email"
    Phone = "phone"
    Firstname = "firstname"
    Lastname = "lastname"
    Address = "address"
    ZipCode = "zipcode"
    City = "city"
    Country = "country"
    Region = "region"
    TimeZone = "timezone"



import time
from functools import wraps

def performance_counter(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = await func(*args, **kwargs)
        duration = time.perf_counter() - start
        print(f"{func.__name__} executed in {duration:.2f} seconds")
        return result

    return wrapper



from typing import Any, Optional, Dict, Tuple, Text, List
from decimal import Decimal, ROUND_DOWN, ROUND_FLOOR
import numpy as np


class NumberUtil:
    @staticmethod
    def is_float(val:str) -> bool:
        try:
            num:float = float(val)
            return True
        except:
            return False


    @staticmethod
    def is_integer(val:str) -> bool:
        try:
            if (isinstance(val, int)):
                return True
            if (isinstance(val, float)):
                return False
            if (val is None):
                return False
            return val.isdigit()
        except:
            return False


    @staticmethod
    def get_safe_float(val:str, precision:Decimal=Decimal('.11111')) -> float:
        """ Get a safe float with a 5 digits rounding precision by default. """
        if ((val is None) | (val == "")):
            return 0
        dec = Decimal(str(val)).quantize(precision, rounding=ROUND_FLOOR)
        if (dec.is_nan()):
            return 0
        return float(dec)


    @staticmethod
    def get_safe_int(val:str, default_val:int=0) -> int:
        if (val is None):
            return default_val
        
        if ((NumberUtil.is_integer(val) == False) &
            (NumberUtil.is_float(val) == False)):
            return default_val

        val = int(val)
        if (np.isnan(val)):
            return default_val

        return val



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from datetime import date, time, datetime, timedelta, timezone



class Token():
    def __init__(self, user_id:str="", expire:datetime=None, data:str="", roles:List[str]=[]) -> None:
        self.userId:str = user_id
        self.expireAt:datetime = expire
        self.data:str = data
        self.roles:List[str] = roles



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from ewoxcore.decorators.serializable import Serializable

T = TypeVar("T")

Factory = Union[Callable[[Any], T], type[T]]

@Serializable
class PagingModel(Generic[T]):
    rows: list[T]

    def __init__(self, data:Optional[dict[str, any]]=None, item_factory:Factory[T]=None) -> None:
        self.rows = []

        if data is None:
            self.numRows: int = 0
            self.skip: int = 0
            self.num: int = 0
        else:
            self.numRows: int = data.get("numRows", 0)
            self.skip: int = data.get("skip", 0)
            self.num: int = data.get("num", 0)
            if (item_factory):
                self.rows = [item_factory(u) for u in data.get("rows", []) if u is not None]


from typing import List
from .key_value_pair import KeyValuePair
from .item_base import ItemBase
from .property_item import PropertyItem


class Item(ItemBase):
    def __init__(self):
        super().__init__()
        self.properties:List[PropertyItem] = []
        self.subItems:List[Item] = []
        self.relations:List[KeyValuePair] = []
        self.externalId:str = ""
        self.isHidden:bool=False

    
    def is_root(self) -> bool:
        if self.parentItemId == "":
            return True
        else:
            return False


    def change_id(self, itemId:str) -> None:
        self.itemId = itemId


    def get_property_item(self, name:str) -> PropertyItem:
        prop:PropertyItem = None
        for p in self.properties:
            if (p.name == name):
                prop = p
                break

        return prop


    def add_property(self, prop:PropertyItem) -> None:
        if (prop is None):
            return

        prop.idx = len(self.properties) + 1
        self.properties.append(prop)


    def get_subitem(self, name:str) -> ItemBase:
        si:ItemBase = None
        for i in self.subItems:
            if (i.name == name):
                si = i
                break

        return si


    def add_subitem(self, item:ItemBase) -> None:
        if (item is None):
            return

        item.parentItemId = self.itemId
        self.subItems.append(item)
    

    def add_relation(self, kv:KeyValuePair) -> None:
        self.relations.append(kv)


    def add_relation(self, key:str, value:str) -> None:
        self.relations.append(KeyValuePair(key, value))



from typing import Any, Optional, Dict, Tuple, Text, List
import os


class DirectoryUtil:
    @staticmethod
    def _filename_without_prefix(file:str) -> str:
        """Splits of a filenames prefix until after the first ``_``."""
        return "_".join(file.split("_")[1:])


    @staticmethod
    def get_filenames(path:str) -> List[str]:
        """Returns all files.

        If the path points to a file, returns the file. This is a recursive
        implementation returning files in any depth of the path."""

        if not isinstance(path, str):
            raise ValueError(
                "`resource_name` must be a string type. "
                "Got `{}` instead".format(type(path))
            )

        if os.path.isfile(path):
            return [path]
        elif os.path.isdir(path):
            results = []
            for base, dirs, files in os.walk(path, followlinks=True):
                # not listing files in cache
                if '__pycache__' in base:
                    continue

                # sort files for same order across runs
                files = sorted(files, key=DirectoryUtil._filename_without_prefix)
    
                # filter out hidden files
                good_files = filter(lambda x: not x.startswith("."), files)

                # filter out files starting with '_'
                good_files = filter(lambda x: not x.startswith("_"), good_files)

                results.extend(os.path.join(base, f) for f in good_files)
            return results
        else:
            raise ValueError(
                "Could not locate the resource '{}'.".format(os.path.abspath(path))
            )


    @staticmethod
    def get_filenames_sorted(path:str):
        items = os.listdir(path)
        sorted_items = sorted(items)

        return sorted_items



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
import json
from queue import Queue
from types import SimpleNamespace
from ewoxcore.utils.class_util import ClassUtil


T = TypeVar("T")


class Configuration():
    def __init__(self) -> None:
        self._settings:List[Dict] = []


    def add(self, settings:Dict) -> Dict:
        if (isinstance(settings, Dict)):
            self._settings.append(settings)


    def _get_dict(self, data:Dict, keys_stack:Queue) -> Optional[Dict]:
        if (keys_stack.empty()):
            return data

        key:str = keys_stack.get()
        if (key in data):
            sub_dict:Dict = data[key]
            return self._get_dict(sub_dict, keys_stack)


    def _get_section_dict(self, key:str) -> Optional[Dict]:
        keys:List[str] = key.split(":")

        for setting in self._settings:
            keys_stack:Queue = Queue()
            for k in keys:
                keys_stack.put(k)

            section:Dict = self._get_dict(setting, keys_stack)
            if (section):
                return section

        return None


    def get_section(self, class_type:T, key:str) -> Optional[T]:
        dict_obj:Dict = self._get_section_dict(key)
        if (dict_obj is None):
            return None

        json_str = json.dumps(dict_obj)
        json_obj = json.loads(json_str, object_hook=lambda d: SimpleNamespace(**d))       
        class_d:T = ClassUtil.merge(class_type(), json_obj)

        return class_d



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from ewoxcore.constants.global_key import GlobalKey

T = TypeVar('T')


def get_service(type:T) -> T:
    globals_dict:Dict[str, Any] = globals()
    if (GlobalKey.GetService not in globals_dict):
        return None

    get_service:Callable[[Type[T]], T] = globals()[GlobalKey.GetService]
    return get_service(type)


def add_service(get_service:Callable[[Type[T]], T]) -> None:
    globals()[GlobalKey.GetService] = get_service



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod

T = TypeVar('T')


class IServiceApp(ABC):
    @abstractmethod
    async def on_start(self) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def on_stop(self) -> None:
        raise NotImplementedError("Implement inhereted method")



import json
from datetime import date, datetime, timedelta
from decimal import *
from enum import *

class DateTimeDecoder(json.JSONDecoder):
    def __init__(self, *args, **kargs):
        json.JSONDecoder.__init__(self, object_hook=self.dict_to_object,
                            *args, **kargs)


    def dict_to_object(self, d):
        if '__type__' not in d:
            return d

        type = d.pop('__type__')
        try:
            dateobj = datetime(**d)
            return dateobj
        except:
            d['__type__'] = type
            return d


    def json_default_format(value):
        try:
            if isinstance(value, datetime):
                return {
                    '__type__': 'datetime',
                    'year': value.year,
                    'month': value.month,
                    'day': value.day,
                    'hour': value.hour,
                    'minute': value.minute,
                    'second': value.second,
                    'microsecond': value.microsecond,
                }
            if isinstance(value, Decimal):
                return float(value)
            if isinstance(value, Enum):
                return value.name
            else:
                return vars(value)
        except Exception as e:
            raise ValueError



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod
from ewoxcore.monitoring.models.log_event import LogEvent

T = TypeVar('T')


class ILogger(ABC):
    @abstractmethod
    def setup(self) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def info(self, message:str, *args, **kwargs) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def warning(self, message:str, *args) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def error(self, message:str, *args) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def debug(self, message:str, *args) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def info_ext(self, event:LogEvent) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def warning_ext(self, event:LogEvent) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def error_ext(self, event:LogEvent) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def debug_ext(self, event:LogEvent) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def log_scoped(self, level: str, message: str, args: Optional[T] = None) -> None:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from datetime import date, datetime, timedelta
from abc import ABC, abstractmethod

T = TypeVar("T")


class ICacheProvider(ABC):
    @abstractmethod
    def activate_expiration(self, timer_interval:int) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def change_use_local(self, use_local:bool) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def set_cache_size(self, size:int) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def get(self, cacheKey:str) -> Any:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def insert(self, cacheKey:str, cacheValue:Any) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def remove(self, cacheKey:str) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def get_all_values(self) -> List[Any]:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def add(self, key:str, value:str, absolute_expiration:timedelta) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def add_expire_at(self, key:str, value:str, expire_at:datetime) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def clear(self) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def add_ext(self, key:str, value:T, absolute_expiration:timedelta) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def add_expire_at_ext(self, key:str, value:T, expire_at:datetime) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def insert_ext(self, key:str, value:T) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def get_ext(self, class_type:T, key:str) -> T:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from ewoxcore.decorators.serializable import Serializable

T = TypeVar("T")

@Serializable
class ListModel():
    def __init__(self, items:list[T]=[]) -> None:
        self.items: list[T] = items



from typing import Any, Optional, Dict, Tuple, Text, List
import asyncio


# https://docs.python.org/3/library/asyncio-task.html#creating-tasks
class TaskUtil:
    """ We keep a reference to the task until it is completed,
        this is to ensure the task does not get garbage collected while running."""
    _tasks:set[asyncio.Task] = set()


    @staticmethod
    def create_task(fn_task:Any) -> asyncio.Task:
        """ Create an asyncio task and keep a reference to it. """
        task:asyncio.Task = asyncio.create_task(fn_task)
        TaskUtil._tasks.add(task)

        # To prevent keeping references to finished tasks forever,
        # make each task remove its own reference from the set after
        # completion.
        task.add_done_callback(TaskUtil._tasks.discard)
        
        return task


    @staticmethod
    async def create_tasks(fn_tasks:List[Any], is_debug:bool=False) -> None:
        """ Create multiple asyncio tasks and wait for them to complete. """
        pending:set[asyncio.Task] = set()
        for fn_task in fn_tasks:
            task:asyncio.Task = TaskUtil.create_task(fn_task)
            pending.add(task)

        if (len(pending) == 0):
            return

        while pending:
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED)

            if (is_debug):
                result = done.pop().result()
                print(result)


    @staticmethod
    def to_thread(fn_task:Any, /, *args, **kwargs) -> asyncio.Task:
        """ Run a function in a separate thread and keep a reference to the task. """
        task = asyncio.create_task(asyncio.to_thread(fn_task, *args, **kwargs))
        TaskUtil._tasks.add(task)

        # To prevent keeping references to finished tasks forever,
        # make each task remove its own reference from the set after
        # completion.
        task.add_done_callback(TaskUtil._tasks.discard)
        
        return task



if __name__ == "__main__":
    async def test_call(name:str, result:int, delay:int):
        print(name)
        await asyncio.sleep(delay)
        return result


    async def main():
        tasks:List[Any] = []
        tasks.append(test_call("TEST1", 1, 20))
        tasks.append(test_call("TEST2", 2, 3))
        tasks.append(test_call("TEST3", 3, 7))
        await TaskUtil.create_tasks(tasks, is_debug=True)

    asyncio.run(main())




from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from datetime import datetime, timedelta, timezone
from abc import ABC, abstractmethod

T = TypeVar('T')


class IAuthorizer(ABC):
    @abstractmethod
    def set_token_timeout(self, timeout: int) -> None:
        """ Set the token timeout in minutes. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def create_token(self, model: T, expires_delta: timedelta=None):
        """ Create a JWT token with the given model and expiration delta. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def decode_token(self, token: str, class_type:T) -> Optional[T]:
        """ Decode the JWT token and return the model. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def get_payload(self, token: str) -> Optional[Dict[str, Any]]:
        """ Decode the JWT token and return the payload. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def is_authorized(self, token:str, required_role:str = "") -> bool:
        """ Check if the token is authorized for the required role. Returns True if authorized, False if not authorized, and False if the token is invalid. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def create_api_token(self, model: T, expires_delta: timedelta=None):
        """ Create an API JWT token with the given model and expiration delta. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def decode_api_token(self, token: str) -> Optional[T]:
        """ Decode the JWT token and return the model. """
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def generate_key_256(self) -> str:
        """ Generate a random 256-bit key for encryption. """
        raise NotImplementedError("Implement inhereted method")



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from datetime import date, time, datetime, timedelta, timezone
from ewoxcore.utils.date_time_util import DatetimeUtil


class LogEvent():
    def __init__(self, correlation_id:str="", user_id:str=None, company_id:str=None, 
                 log_level:int=1, message:str="", event_name:str="", event_type:str="", 
#                 data_class:str ="", data:str="", is_encoded:bool=False, created_at:datetime=datetime.now(tz=timezone.utc)()) -> None:
                 data_class:str ="", data:str="", is_encoded:bool=False, created_at:datetime=datetime.now(tz=timezone.utc)) -> None:
        self.correlationId:str = correlation_id
        self.userId:str = user_id
        self.companyId:str = company_id
        self.logLevel:int = log_level
        self.message:str = str(message)
        self.eventName:str = event_name
        self.eventType:str = event_type
        self.dataClass:str = data_class
        self.data:str = data
        self.isEncoded:bool = is_encoded # Base64 encoded data.
        self.createdAt:datetime = created_at



from enum import IntEnum


class IntervalType(IntEnum):
    FireAndForget = 0   # will run immediately
    FireAndForgetAt = 1 # will run based on start date
    Year = 2
    Month = 3
    Week = 4
    Day = 5
    Hour = 6
    Minute = 7
    Second = 8



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union, TypeVar, Type
from types import SimpleNamespace
import base64
import jsonpickle
import json
import re
from ewoxcore.utils.class_util import ClassUtil
from ewoxcore.utils.dictionary_util import DictionaryUtil

T = TypeVar("T")

"""
Use Humps for converting between snake case, camelcase and pascalcase
https://humps.readthedocs.io/en/latest/
"""
class JsonUtil:

    @staticmethod
    def serializeJson64(val, keep_object:bool=False) -> str:
        if (val is None):
            return ""
        
        # unpickable must be set to true, otherwise deserialize to an object will not work.
        json_str = jsonpickle.encode(val, unpicklable=keep_object, make_refs=False)
    
        json64:str = base64.standard_b64encode(json_str.encode()).decode()
    
        return json64


    @staticmethod
    def deserializeJson64(val:str, to_object:bool=True):
        """ Deperected, use: deserialize_json64"""
        if ((val == "") | (val is None)):
            return None
        json64 = base64.standard_b64decode(val)
        json_str = json64.decode()
        decObj = jsonpickle.decode(json_str)
        if ((to_object) and (type(decObj) == type({}))):
            decObj = JsonUtil.deserialize_object(json_str)
            return decObj

        return decObj


    @staticmethod
    def deserialize_json64(class_type:T, val:str, to_object:bool=True, force_merge:bool=True):
        if ((val == "") | (val is None)):
            return None
        json64 = base64.standard_b64decode(val)
        json_str = json64.decode()
        decObj = jsonpickle.decode(json_str)
        if ((to_object) and (type(decObj) == type({}))):
            decObj = JsonUtil.deserialize_object(json_str)
            if (force_merge):
                decObj = ClassUtil.merge_from_type(class_type, decObj)
            else:
                decObj = ClassUtil.merge(class_type(), decObj)
            return decObj

        if ((to_object) and (force_merge)):
            decObj = ClassUtil.merge(class_type(), decObj)
            return decObj

        return decObj
    

    @staticmethod
    def serialize(val, keep_object:bool=False) -> str:
        if (val is None):
            return ""
        
        json_str = jsonpickle.encode(val, unpicklable=keep_object)
        
        return json_str


    @staticmethod
    def serialize_list(val) -> str:
        json_str = json.dumps(val)
        return json_str


    @staticmethod
    def deserialize_object(val:str):
        if (val == ""):
            return None
        decObj = json.loads(val, object_hook=lambda d: SimpleNamespace(**d))
        return decObj


    @staticmethod
    def deserialize_gen_object(class_type:T, val:str):
        if (val == ""):
            return None
        decObj = json.loads(val, object_hook=lambda d: SimpleNamespace(**d))       
        decObj = ClassUtil.merge(class_type(), decObj)

        return decObj


    @staticmethod
    def deserialize(val:str):
        if (val == ""):
            return None
        decObj = jsonpickle.decode(val)
        return decObj


    @staticmethod
    def write_dict(file_name:str, data:Dict):
        with open(file_name, 'w') as outfile:
            json.dump(data, outfile, ensure_ascii=False, indent=4)


    @staticmethod
    def read_dict(file_name:str):
        with open(file_name) as json_file:
            data = json.load(json_file)
            return data


    @staticmethod
    def write(file_name:str, model:Any):
        json_data = json.dumps(model, default=lambda o: o.__dict__, indent=4)
        with open(file_name, 'w') as outfile:
            outfile.write(json_data)


    @staticmethod
    def write_utf8(file_name:str, model:Any):
        json_data = json.dumps(model, default=lambda o: o.__dict__, indent=4, ensure_ascii=False)
        with open(file_name, 'w', encoding='utf8') as outfile:
            outfile.write(json_data)


    @staticmethod
    def write_text(file_name:str, text:str):
        with open(file_name, 'w') as outfile:
            outfile.write(text)


    @staticmethod
    def is_base64(val:str) -> bool:
        x = re.search("^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$", val)
        if (x):
            return True
        return False


    @staticmethod
    def decode_json(val:str) -> str:
        json64 = base64.standard_b64decode(val)
        json_str = json64.decode()
        return json_str


    @staticmethod
    def encodeBase64(json_str:Any) -> str:
        json64:str = base64.standard_b64encode(json_str.encode()).decode()
        return json64


    @staticmethod
    def get_value(json_str:str, name:str, default:Any) -> Any:
        json_dict:Dict = json.loads(json_str)
        return DictionaryUtil.get(json_dict, name, default)



from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from datetime import date, datetime, timedelta, timezone


class MonitoringEvent():
    def __init__(self, service_id:str="", serivce_name:str="", event_name:str="", event_data:str="") -> None:
      self.service_id:str = service_id
      self.serivce_name:str = serivce_name
      self.event_name:str = event_name
      self.event_data:str = event_data
      self.created_at:datetime = datetime.now(timezone.utc)



from typing import Any, Optional, Dict, Tuple, Text, List


class ListUtil:
    def __init__(self):
        pass
    

    @staticmethod
    def split(str_list:List[str], separator:str) -> List[List[str]]:
        splitted:List[List[str]] = []
        tmp_list:List[str] = []
        for i, s in enumerate(str_list):
            if (s != separator):
                tmp_list.append(s)
            else:
                splitted.append(tmp_list)
                tmp_list = []

        if (len(tmp_list) > 0):
            splitted.append(tmp_list)

        return splitted


    @staticmethod
    def to_string(str_list:List[str], separator:str = ",") -> str:
        s:str = separator.join(str_list)
        return s



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from datetime import date, datetime, timedelta
from ewoxcore.monitoring.models.log_event import LogEvent
from ewoxcore.utils.json_util import JsonUtil


class LogEventParser():
    @staticmethod
    def parse(response:Any) -> Tuple[int, List[LogEvent]]:
        m_json:Any = JsonUtil.serialize(response)
        m_dec:Any = JsonUtil.deserialize_object(m_json)

        total:int = m_dec.hits.total.value

        items:List[LogEvent] = []
        for qe in m_dec.hits.hits:
            qe_data:LogEvent = qe._source
            event = LogEvent(
                correlation_id=qe_data.correlationId,
                user_id=qe_data.userId,
                company_id=qe_data.companyId,
                log_level=qe_data.logLevel,
                message=qe_data.message,
                event_name=qe_data.eventName,
                event_type=qe_data.eventType,
                data_class=qe_data.dataClass,
                data=qe_data.data,
                is_encoded=qe_data.isEncoded,
                created_at=qe_data.createdAt
            )
            items.append(event)

        return (total, items)
    

    @staticmethod
    def parse_aggregate(response:Any) -> Tuple[int, List[str]]:
        m_json:Any = JsonUtil.serialize(response)
        m_dec:Any = JsonUtil.deserialize_object(m_json)

        total:int = m_dec.hits.total.value

        items:List[str] = []
        for qe in m_dec.aggregations.aggsTerm.buckets:
            items.append(qe.key)

        return (total, items)


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar, Type
from abc import ABC, abstractmethod
from ewoxcore.service.interfaces.iservice_collection import IServiceCollection

T = TypeVar('T')


class IStartup(ABC):
    @abstractmethod
    def configure_services(self, service:IServiceCollection, get_service:Callable[[Type[T]], T], use_unittest:bool=False) -> None:
        raise NotImplementedError("Implement inhereted method")


#    @abstractmethod
#    def setup(self, get_service:Callable[[Type[T]], T]) -> None:
#        """ Setup is being called after all service configurations. """
#        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    async def setup_async(self, get_service:Callable[[Type[T]], T]) -> None:
        """ Setup is being called after all service configurations. """
        pass


from typing import Any, Callable, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from abc import ABC, abstractmethod


class ILocalizer(ABC):
    @abstractmethod
    def register(self, language_code:str, language_two_letter:str) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def setup(self, path:str="./data/translations", default_language:str="en-GB") -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def get(self, key:str, language_code:str="en-GB") -> str:
        raise NotImplementedError("Implement inhereted method")



from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union
from abc import ABC, abstractmethod
from ewoxcore.items.item import Item


class IMemoryHandle(ABC):
    @abstractmethod
    def add(self, variable:str, inner_state:str) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def remove(self, variable:str, inner_state:str) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def get_inner_states(self, variable:str) -> List[str]:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def contains(self, variable:str) -> bool:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def check(self, variable:str, inner_state:str) -> bool:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def add_item(self, variable:str, item:Item) -> None:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def get_items_keys(self) -> List[str]:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def get_items(self, variable:str) -> List[Item]:
        raise NotImplementedError("Implement inhereted method")


    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError("Implement inhereted method")



import types
import importlib
from typing import Any, List, Dict, Text, Optional, Tuple, Union


class LazyLoader(types.ModuleType):
    def __init__(self, module_name:str, submod_name=None):
        self._module_name = '{}{}'.format(
        module_name,
        submod_name and '.{}'.format(submod_name) or ''
        )
        self._mod = None
        super(LazyLoader, self).__init__(self._module_name)

    def _load(self):
        if self._mod is None:
            self._mod = importlib.import_module(self._module_name)
        return self._mod

    def __getattr__(self, attrb):
        return getattr(self._load(), attrb)
    
    def __dir__(self):
        return dir(self._load())



from typing import Any, Optional, Dict, Tuple, Text, List
import validators
import re


class StringUtil:

    @staticmethod
    def get_safe_name(name:str) -> str:
        safe_name:str = re.sub("[^A-Za-z0-9--_]", '', name)
        return safe_name


    @staticmethod
    def strip_html(text:str) -> str:
        tag_re = re.compile(r'(<!--.*?-->|<[^>]*>)')
        no_tags = tag_re.sub('', text)
        return no_tags


    @staticmethod
    def is_email(input:str) -> bool:
        try:
            res:bool = validators.email(input)
            return True if (res) else False
        except:
            return False


    @staticmethod
    def extract_tags(text:str, begin:str="<%", end:str="%>") -> List[Tuple[str, str, str]]:
        return re.findall('('+begin+')(.+?)('+end+')', text, re.DOTALL)


    @staticmethod
    def extract_tag_keys(text:str, begin:str="<%", end:str="%>") -> List[str]:
        tups:List[Tuple[str, str, str]] = StringUtil.extract_tags(text, begin, end)
        keys:List[str] = []
        for tb, key, te in tups:
            if (key not in keys):
                keys.append(key)
        return keys


    @staticmethod
    def title_capitalized(text:str) -> str:
        """ Returns only the first letter in upper case. """
        if (text == ""):
            return ""
        return text[0].upper() + text[1:]


    @staticmethod
    def get_safe_string(val:str) -> str:
        if (val is None):
            return ""
        return str(val)


    @staticmethod
    def strip_end(val:str, ch:str) -> str:
        res:List[str] = val.split(ch)
        return res[0]


    @staticmethod
    def capitalize_camelcase_words(val:str) -> str:
        def camel_case_segment(segment):
            parts = re.split(r'[_\-]', segment)
            if not parts:
                return ''
            # Join camel case inside the segment but keep casing (just capitalize first char of each part except first)
            camel_cased = parts[0] + ''.join(p[0].upper() + p[1:] if len(p) > 1 else p.upper() for p in parts[1:])
            return camel_cased

        def process_word(word):
            camel_word = camel_case_segment(word)
            # Capitalize first letter of the whole word
            return camel_word[0].upper() + camel_word[1:] if camel_word else ''

        # Split by spaces but keep spaces
        tokens = re.split(r'(\s+)', val)
        result = [process_word(token) if token.strip() else token for token in tokens]
        return ''.join(result)



from typing import Any, Optional, Dict, Tuple, get_args, get_origin, get_type_hints, Union, ForwardRef, TypeVar
from collections.abc import Mapping
from argparse import Namespace
from types import SimpleNamespace
import sys
from ewoxcore.utils.dictionary_util import DictionaryUtil

_PRIMITIVES = {str, int, float, bool, bytes}

T = TypeVar("T")


class ClassUtil:
    @staticmethod
    def merge(class_a:Any, class_b:Any) -> Any:
        if hasattr(class_a, "_merge_classes") and callable(getattr(class_a, "_merge_classes", None)):
            class_a._merge_classes(class_b.__dict__)
        else:
            class_a.__dict__.update(class_b.__dict__)

        return class_a


    @staticmethod
    def merge_from_type(class_type:T, class_b:Any) -> Any:
        model:T = None
        class_b:dict[str, Any] = DictionaryUtil.normalize(class_b)
        if (ClassUtil.is_namespace(class_b)):
            model = class_type(class_b.__dict__)
        elif (ClassUtil.is_dict(class_b)):
            model = class_type(class_b)
        else:
            raise Exception("class_b must be a Namespace or a Dict")

        return model


    @staticmethod
    def get_class_name(cls:Any) -> str:
        if cls is None:
            return ""
        return cls.__name__ if isinstance(cls, type) else type(cls).__name__


    @staticmethod
    def get_class_instance_name(cls:Any) -> str:
       return cls.__class__.__name__ if cls is not None else ""


    @staticmethod
    def is_namespace(val:Any) -> bool:
        return isinstance(val, (Namespace, SimpleNamespace))


    @staticmethod
    def is_dict(val:Any) -> bool:
        return isinstance(val, Mapping)

    # @staticmethod
    # def get_class_names(model: Any) -> List[str]:
    #     class_names = set()

    #     # Add class name of the model itself
    #     class_names.add(type(model).__name__)

    #     # Get attributes of the model
    #     try:
    #         attributes = vars(model)
    #     except TypeError:
    #         # model has no __dict__ (e.g. primitive)
    #         return List(class_names)

    #     for attr_name, attr_value in attributes.items():
    #         if attr_value is None:
    #             continue

    #         # Check if it's a list and has elements
    #         if isinstance(attr_value, list) and len(attr_value) > 0:
    #             first_elem = attr_value[0]
    #             elem_class_name = type(first_elem).__name__
    #             if elem_class_name not in ["dict", "list", "str", "int", "float", "bool"]:
    #                 class_names.add(elem_class_name)

    #         # Check if it's another object (not primitive, not list)
    #         elif hasattr(attr_value, "__dict__"):
    #             value_class_name = type(attr_value).__name__
    #             if value_class_name not in ["dict", "list", "str", "int", "float", "bool"]:
    #                 class_names.add(value_class_name)

    #     return list(class_names)

                # qualname:str = tp.__qualname__
                # print(f"{type(tp).__module__}.{type(tp).__qualname__}")


    # @staticmethod
    # def get_class_names_old(model: Any) -> List[str]:
    #     names: Set[str] = set()
    #     seen: Set[type] = set()

    #     def add_type(tp: Any):
    #         if isinstance(tp, type) and tp not in _PRIMITIVES and tp not in {list, dict, tuple, set}:
    #             names.add(tp.__name__)
    #             return True
    #         return False

    #     def walk_type(tp: Any, queue: List[type]):
    #         origin = get_origin(tp)
    #         if origin is None:
    #             if add_type(tp):
    #                 queue.append(tp)
    #             return
    #         if origin is Union:
    #             for arg in get_args(tp):
    #                 if arg is type(None):
    #                     continue
    #                 walk_type(arg, queue)
    #             return

    #         # List[T], Dict[K,V], etc.
    #         for arg in get_args(tp):
    #             walk_type(arg, queue)


    #     def inspect_class(cls: type, queue: List[type]):
    #         if cls in seen:
    #             return
    #         seen.add(cls)
    #         mod_globals = vars(sys.modules.get(cls.__module__, object()))
    #         hints = {}
    #         try:
    #             hints = get_type_hints(cls, globalns=mod_globals, localns=vars(cls))
    #         except Exception:
    #             hints = getattr(cls, "__annotations__", {}) or {}
    #         for tp in hints.values():
    #             walk_type(tp, queue)

    #     # seed with the model’s own type
    #     names.add(type(model).__name__)
    #     queue: List[type] = [type(model)]

    #     # Instance values to recover types from populated lists/objects
    #     try:
    #         for val in vars(model).values():
    #             if val is None:
    #                 continue
    #             if isinstance(val, list) and val:
    #                 if add_type(type(val[0])): queue.append(type(val[0]))
    #             elif not isinstance(val, (str, int, float, bool, bytes, list, dict, tuple, set)):
    #                 if add_type(type(val)): queue.append(type(val))
    #     except TypeError:
    #         pass

    #     while queue:
    #         inspect_class(queue.pop(), queue)

    #     return sorted(names)


    @staticmethod
    def get_class_names(model: Any) -> list[str]:
        names: set[str] = set()
        seen_types: set[type] = set()
        seen_ids: set[int] = set()  # for instance graph cycle protection

        def add_type(tp: Any) -> bool:
            if isinstance(tp, type) and tp not in _PRIMITIVES and tp not in {list, dict, tuple, set}:
                names.add(tp.__name__)
                return True
            return False

        # ----- Static/type-hint walk -----
        def walk_type(tp: Any, tqueue: list[type]):
            origin = get_origin(tp)
            if origin is None:
                if add_type(tp):
                    tqueue.append(tp)
                return
            if origin is Union:
                for arg in get_args(tp):
                    if arg is type(None):
                        continue
                    walk_type(arg, tqueue)
                return
            # e.g. List[T], Dict[K, V], Tuple[…], etc.
            for arg in get_args(tp):
                walk_type(arg, tqueue)

        def inspect_class(cls: type, tqueue: list[type]):
            if cls in seen_types:
                return
            seen_types.add(cls)
            mod_globals = vars(sys.modules.get(cls.__module__, object()))
            try:
                hints = get_type_hints(cls, globalns=mod_globals, localns=vars(cls))
            except Exception:
                hints = getattr(cls, "__annotations__", {}) or {}
            for tp in hints.values():
                walk_type(tp, tqueue)

        # ----- Dynamic/instance walk (RECURSIVE) -----
        def walk_instance(obj: Any):
            oid = id(obj)
            if oid in seen_ids:
                return
            seen_ids.add(oid)

            # record this object's concrete type
            add_type(type(obj))

            # containers
            if isinstance(obj, (list, tuple, set)):
                for item in obj:
                    walk_instance(item)
                return
            if isinstance(obj, dict):
                for k, v in obj.items():
                    walk_instance(k)
                    walk_instance(v)
                return

            # primitives stop here
            if isinstance(obj, tuple(_PRIMITIVES)):
                return

            # generic objects: traverse their attributes
            try:
                for val in vars(obj).values():
                    if val is not None:
                        walk_instance(val)
            except TypeError:
                # objects without __dict__
                pass

        # seed with the model’s own type
        add_type(type(model))

        # Walk the instance graph to capture runtime subclasses based on inheritance
        walk_instance(model)

        # Walk the type graph to include annotated types you might not have populated yet
        tqueue: list[type] = list({type(model), *[type(v) for v in vars(model).values()
                                                if not isinstance(v, tuple(_PRIMITIVES | {list, dict, tuple, set}))]})
        while tqueue:
            inspect_class(tqueue.pop(), tqueue)

        return sorted(names)


from typing import Any, Callable, Generic, List, Dict, Text, Optional, Tuple, Union
from queue import LifoQueue
import threading
import copy


class Stack():
    def __init__(self, max_size:int=0) -> None:
        self._collection:LifoQueue = LifoQueue(maxsize=max_size)
        self._lock = threading.Lock()


    def get_length(self) -> int:
        return self._collection.qsize()


    def push(self, item:Any) -> None:
        with self._lock:
            self._collection.put(item)


    def pop(self, wait:bool=False) -> Optional[Any]:
        with self._lock:
            if (self._collection.empty()):
                return None
            return self._collection.get() if (wait) else self._collection.get_nowait()



from typing import Any, List, Dict, Text, Optional, Tuple, Union
from datetime import datetime, timezone
import uuid

class ServiceIdentifier():
    def __init__(self, name:str) -> None:
        self.id:str = str(uuid.uuid4())
        self.name:str = name



from typing import Any, List, Optional, Dict, Tuple, Text
from datetime import date, time, datetime, timedelta, timezone


class StopWatch():
    def __init__(self) -> None:
        self._is_running:bool = False
        self._start_at:time = None
        self._stop_at:time = None


    def start(self) -> datetime:
        self._start_at = datetime.now()
        return self._start_at


    def stop(self) -> datetime:
        self._stop_at = datetime.now()
        return self._stop_at


    def elapsed(self) -> timedelta:
        delta:timedelta = self._stop_at - self._start_at
        return delta


if __name__ == "__main__":
    sw = StopWatch()
    sw.start()
    sw.stop()
    sw.elapsed()



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable
from datetime import date, datetime, timedelta
from .item import Item
from ewoxcore.utils.json_util import JsonUtil


class ItemConverter():
    @staticmethod
    def convert_to_json(item:Item) -> str:
        if (item is None):
            return "{}"

        json_dict:Dict = dict()
        for prop in item.properties:
            if (prop.name):
                json_dict[prop.name] = prop.value

        json_str:str = JsonUtil.serialize(json_dict, keep_object=False)
        return json_str



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable, TypeVar
from datetime import date, datetime, timedelta
from ewoxcore.monitoring.models.log_event import LogEvent


class LogEvents():
    def __init__(self) -> None:
        self.events:List[LogEvent] = []



from typing import Any, List, Dict, Text, Optional, Tuple, Union, Awaitable
import time
import threading


class TimerExt(threading.Thread):
    def __init__(self, callback, interval_seconds:int, activate_immediately:bool):
        super().__init__()
        self.callback = callback
        self._interval_seconds = interval_seconds
        self._running:bool = True
        self._active:bool = True if (activate_immediately) else False


    def run(self):
        while self._running:
            if self._active:
                for x in range(0,self._interval_seconds):
                    if self._active:
                        time.sleep(1)
                self.callback()
            else:
                time.sleep(float(0.1))


    def stop(self):
        self._active = False
        self._running = False


    def activate(self):
        self._active = True


    def deactivate(self):
        self._active = False


    def setactive(self, active):
        self._active = active